import sys

from os import environ
from pathlib import Path
from random import choice, choices

from iterfzf import iterfzf
from pydantic import BaseModel
from rich.prompt import Confirm, IntPrompt, Prompt

from .console import console
from .constants import (
    ABC_PARAMS,
    ABF_PARAMS,
    ABFB_PARAMS,
    BODYWEIGHT_FACTORS,
    DFW_PARAMS,
    BTB_PARAMS,
    # EXERCISES,
    FZF_DEFAULT_OPTS,
    HUMAN_MOVEMENTS,
    IRON_CARDIO_PARAMS,
    POUNDS_TO_KILOS_RATE,
    PW_PARAMS,
    ROP_PARAMS,
    SUGGESTION,
    THE_GIANT_PARAMS,
    THE_WOLF_PARAMS,
    WARNING,
    WORKOUT_GENERATOR_PARAMS,
    KETTLEBELLS_DB,
)
from .database import read_database, write_database


class Exercise(BaseModel):
    """Exercise object.

    Attributes:
        name: str
        load: int
        sets: int
        reps: int
    """

    name: str
    load: int
    sets: int
    reps: int


class Workout(BaseModel):
    """Workout object.

    Attributes:
        workout_type: str
        variation: str
        time: int
        units: str
        bodyweight: int
        exercises: list[Exercise]
    """

    workout_type: str
    variation: str
    time: int
    units: str
    bodyweight: int
    exercises: list[Exercise]

    def display_workout(self) -> None:
        """Print a workout to the console."""
        console.print(self.workout_type.upper())
        console.print("=" * len(self.workout_type), style="green")
        display_params = [("Variation", self.variation), ("Time", f"{self.time} mins")]
        if self.workout_type in ["iron cardio", "armor building complex"]:
            display_params.append(("Load", f"{self.exercises[0].load} {self.units}"))
            swings = [e for e in self.exercises if "Swing" in e.name]
            if swings:
                display_params.append(("Kettlebell Swing", swings[0].reps))
        elif self.workout_type == "back to basics":
            second_block = self.exercises[-1]
            display_params.append(
                ("Clean and Press Load", f"{self.exercises[0].load} {self.units}")
            )
            display_params.append(
                (f"{second_block.name} Load", f"{second_block.load} {self.units}")
            )
        else:
            exercises = []
            for exercise in self.exercises:
                exercises.append(
                    (
                        exercise.name,
                        f"{exercise.sets} X {exercise.reps} with {exercise.load} {self.units}",
                    )
                )
            _print_helper(display_params)
            console.print("Exercises")
            console.print("---------", style="green")
            _print_helper(exercises)
            return
        _print_helper(display_params)

    def calc_workout_stats(self) -> dict:
        """Calculate the stats for a given workout.

        Returns:
            A dict containing total weight moved, reps, weight density and rep density.
        """
        total_reps = 0
        weight_moved = 0
        for exercise in self.exercises:
            if "double" in exercise.name.lower():
                load = exercise.load * 2
            else:
                load = exercise.load
            if "clean and press" in exercise.name.lower():
                reps = 2 * exercise.reps * exercise.sets
            else:
                reps = exercise.reps * exercise.sets
            total_reps += reps
            weight_moved += reps * load
        stats = {
            "weight moved": weight_moved,
            "reps": total_reps,
            "weight density": round(weight_moved / self.time, 1),
            "rep density": round(total_reps / self.time, 1),
        }
        return stats

    def display_workout_stats(self) -> None:
        """Prints the stats for a given workout.

        Returns:
            None
        """
        stats = self.calc_workout_stats()
        console.print("Workout Stats")
        console.print("=============", style="green")
        stats_to_print = [
            ("Weight Moved", f"{stats.get('weight moved'):,} {self.units}"),
            ("Total Reps", f"{stats.get('reps')}"),
            ("Weight Density", f"{stats.get('weight density')} kg/min"),
            ("Rep Density", f"{stats.get('rep density')} reps/min"),
        ]
        _print_helper(stats_to_print)


def random_ic_or_abc(db_path: Path, workout_type: str) -> Workout:
    """Create a random workout based on workout_type.

    Args:
        db_path:  The Path to the database.
        workout_type:  Either ic or abc.
    Returns:
        A Workout object with randomly generated parameters.
    """
    data = read_database(db_path)
    loads = data["loads"]
    workout_type, workout_params = _get_workout_params(workout_type)
    bells = choices(
        population=tuple(workout_params["bells"].keys()),
        weights=tuple(workout_params["bells"].values()),
    )[0]
    if bells == "Double Bells":
        variation = choices(
            population=tuple(workout_params["doublebell variations"].keys()),
            weights=tuple(workout_params["doublebell variations"].values()),
        )[0]
    elif bells == "Single Bell":
        variation = choices(
            population=tuple(workout_params["singlebell variations"].keys()),
            weights=tuple(workout_params["singlebell variations"].values()),
        )[0]
    time = choices(
        population=tuple(workout_params["times"].keys()),
        weights=tuple(workout_params["times"].values()),
    )[0]
    load = choices(
        population=tuple(workout_params["loads"].keys()),
        weights=tuple(workout_params["loads"].values()),
    )[0]
    swings = choices(
        population=tuple(workout_params["swings"].keys()),
        weights=tuple(workout_params["swings"].values()),
    )[0]
    load = loads[load]
    units = loads["units"]
    exercises = []
    for exercise in workout_params["exercises"][variation]:
        exercises.append(
            Exercise(name=exercise[0], load=load, sets=0, reps=exercise[1])
        )

    if swings:
        swings = Exercise(
            name="Kettlebell Swing",
            load=load,
            sets=1,
            reps=choice(range(50, 160, 10)),
        )
        exercises.append(swings)
    return Workout(
        workout_type=workout_type,
        variation=variation,
        time=time,
        units=units,
        bodyweight=data["loads"]["bodyweight"],
        exercises=exercises,
    )


def create_ic_or_abc(db_path: Path, workout_type: str) -> Workout:
    """Create an ic or abc Workout.

    Args:
        db_path: The Path object to the database.
        workout_type: The workout type either 'ic' or 'abc'

    Returns:
        A Workout object created by the user.
    """
    data = read_database(db_path)
    bodyweight = data["loads"]["bodyweight"]
    units = data["loads"]["units"]
    workout_type, workout_params = _get_workout_params(workout_type)
    bells = _get_options(workout_params["bells"])
    if bells == "Double Bells":
        variation = _get_options(workout_params["doublebell variations"])
    elif bells == "Single Bell":
        variation = _get_options(workout_params["singlebell variations"])
    time = IntPrompt.ask("How long was your workout (mins)")
    load = IntPrompt.ask(f"What weight did you use ({units})")
    sets = IntPrompt.ask("How many sets did you complete?")
    exercises = []
    if "Traveling 2s" in variation:
        sets = sets + sets // 3
    for exercise, reps in workout_params["exercises"][variation]:
        if bells == "Single Bell" and exercise == "Pullup":
            sets = sets // 2
        if exercise == "Pullup":
            load = _add_bodyweight_factor(bodyweight, exercise)
        exercises.append(
            Exercise(
                name=exercise,
                load=load,
                sets=sets,
                reps=reps,
            )
        )
    if Confirm.ask("Did you swing"):
        swings = IntPrompt.ask("How many swings")
        load = IntPrompt.ask(f"What weight did you use ({units})")
        exercises.append(
            Exercise(
                name="Kettlebell Swing",
                load=load,
                sets=1,
                reps=swings,
            )
        )
    return Workout(
        workout_type=workout_type,
        variation=variation,
        time=time,
        units=units,
        bodyweight=bodyweight,
        exercises=exercises,
    )


def set_loads() -> dict:
    """Creates a dictionary containing the units and kettlebell weights and body weight
    of the user.

    Returns:
        A dict of the loads set by the user.

    """
    while True:
        units = _get_units()
        bodyweight = IntPrompt.ask("Bodyweight")
        console.print("Enter the weight for the...")
        light_load = IntPrompt.ask("Light kettlebell")
        medium_load = IntPrompt.ask("Medium kettlebell")
        heavy_load = IntPrompt.ask("Heavy kettlebell")
        loads = {
            "units": units,
            "bodyweight": bodyweight,
            "light load": light_load,
            "medium load": medium_load,
            "heavy load": heavy_load,
        }
        console.clear()
        for label, value in loads.items():
            console.print(f"{label.title()}: {value}")
        if Confirm.ask(
            "Are these loads correct? If you confirm, they will be used to generate workouts."
        ):
            return loads


def set_program_loads(loads: dict) -> dict:
    """Adds a program and load to the loads dictionary.

    Returns:
        The loads dictionary with an added item of the program and load.

    """
    program = Prompt.ask("Enter the name of the program").lower()
    program_load = IntPrompt.ask(f"Enter the weight for the {program.title()}")
    loads[program] = program_load
    return loads


def create_btb_workout(db_path: Path) -> Workout:
    """Create a back to basics kettelbell workout.

    Args:
        db_path: The Path to the database.

    Returns:
        A workout object.

    """
    data = read_database(db_path)
    bodyweight = data["loads"]["bodyweight"]
    units = data["loads"]["units"]
    workout_type, workout_params = _get_workout_params("btb")
    variation = _get_options(workout_params)
    if "Squat" in variation:
        second_block = "double kettlebell front squats"
    else:
        second_block = "kettlebell snatch"
    time = IntPrompt.ask("Workout duration (mins)")
    console.print("Enter the weight used for the...")
    c_and_p_load = IntPrompt.ask(f"clean and press ({units})")
    second_block_load = IntPrompt.ask(f"{second_block} ({units})")
    c_and_p = workout_params[variation]["exercises"][0]
    second_block_exercise = workout_params[variation]["exercises"][1]
    exercises = [
        Exercise(
            name=c_and_p["name"],
            load=c_and_p_load,
            sets=c_and_p["sets"],
            reps=c_and_p["reps"],
        ),
        Exercise(
            name=second_block_exercise["name"],
            load=second_block_load,
            sets=second_block_exercise["sets"],
            reps=second_block_exercise["reps"],
        ),
    ]
    return Workout(
        workout_type=workout_type,
        variation=variation,
        time=time,
        units=units,
        bodyweight=bodyweight,
        exercises=exercises,
    )


def create_perfect_workout(db_path: Path) -> Workout:
    """Create a Dan John perfect workout.

    Args:
        db_path: The Path to the database.

    Returns:
        A workout object.

    """
    data = read_database(db_path)
    bodyweight = data["loads"]["bodyweight"]
    units = data["loads"]["units"]
    workout_type, workout_params = _get_workout_params("pw")
    variation = _get_options(workout_params)
    time = IntPrompt.ask("Workout duration (mins)")
    console.print("Enter the weight for the...")
    exercises = []
    no_load_exercises = ["Broomstick", "Hanging"]
    for exercise in workout_params[variation]["exercises"]:
        if not any(x in exercise["name"] for x in no_load_exercises):
            exercise["load"] = IntPrompt.ask(f"  {exercise['name']}")
        exercises.append(Exercise(**exercise))
    return Workout(
        workout_type=workout_type,
        variation=variation,
        time=time,
        units=units,
        bodyweight=bodyweight,
        exercises=exercises,
    )


def create_easy_strength_workout(db_path: Path, workout_type: str) -> Workout:
    """Create an Easy Strength workout.

    Args:
        db_path: The Path to the database.

    Returns:
        A workout object.

    """
    data = read_database(db_path)
    bodyweight = data["loads"]["bodyweight"]
    units = data["loads"]["units"]
    workout_type, workout_params = _get_workout_params(workout_type)
    variation = _get_options(workout_params)
    time = IntPrompt.ask("Workout duration (mins)")
    exercises = []
    for exercise in workout_params[variation]["exercises"]:
        exercise = Exercise(**exercise)
        exercise.load += _add_bodyweight_factor(bodyweight, exercise.name)
        exercises.append(exercise)
    variation = IntPrompt.ask("What day of Easy Strength is this")
    return Workout(
        workout_type=workout_type,
        variation=f"{variation:0>2}",
        time=time,
        units=units,
        bodyweight=bodyweight,
        exercises=exercises,
    )


def create_time_based_workout(db_path: Path, workout_type: str) -> Workout:
    """Create a workout based on a given amount of time. E.g., The Giant.

    Args:
        db_path: The path to the database.
        workout_type: The name of the program.

    Returns:
        The workout object for a given day of a program.
    """
    data = read_database(db_path)
    bodyweight = data["loads"]["bodyweight"]
    units = data["loads"]["units"]
    workout_type, workout_params = _get_workout_params(workout_type)
    program = _get_options(workout_params)
    try:
        load = data["loads"][workout_type]
    except KeyError:
        console.print(
            f"Could not find load for {workout_type} in the database.", WARNING
        )
        console.print(
            "Try running [underline]kettlebells setloads -p[/underline]", SUGGESTION
        )
    week = Prompt.ask("Enter the week")
    day = Prompt.ask("Enter the day")
    variation = f"W{week}D{day}"
    time = IntPrompt.ask("Workout duration (mins)")
    sets = IntPrompt.ask("Number of sets")
    exercises = []
    for exercise in workout_params[program][variation]:
        if isinstance(exercise["reps"], list):
            exercise["reps"] = sum(exercise["reps"])
        exercise["sets"] = sets
        exercise["load"] = load
        exercises.append(Exercise(**exercise))
    if Confirm.ask("Did you do any auxiliary exercises"):
        auxiliary_exercises = add_exercises(db_path, bodyweight, units)
        exercises.extend(auxiliary_exercises)

    return Workout(
        workout_type=program,
        variation=variation,
        time=time,
        units=units,
        bodyweight=bodyweight,
        exercises=exercises,
    )


def create_set_based_workout(db_path: Path, workout_type: str) -> Workout:
    """Create a workout based on a given number of sets.

    Args:
        db_path: The path to the database.
        workout_type: The name of the program.

    Returns:
        The workout object for a given day of a program.
    """
    data = read_database(db_path)
    bodyweight = data["loads"]["bodyweight"]
    units = data["loads"]["units"]
    workout_type, workout_params = _get_workout_params(workout_type)
    program = _get_options(workout_params)
    try:
        load = data["loads"][workout_type]
    except KeyError:
        console.print(
            f"Could not find load for {workout_type} in the database.", style=WARNING
        )
        console.print(
            "Try running [underline]kettlebells setloads -p[/underline]",
            style=SUGGESTION,
        )
        sys.exit()
    week = Prompt.ask("Enter the week")
    day = Prompt.ask("Enter the day")
    variation = f"W{week}D{day}"
    time = IntPrompt.ask("How long was your workout (mins)")
    exercises = []
    for exercise in workout_params[program][variation]:
        exercise["load"] = load
        exercises.append(Exercise(**exercise))
    return Workout(
        workout_type=program,
        variation=variation,
        time=time,
        units=units,
        bodyweight=bodyweight,
        exercises=exercises,
    )


def create_workout_generator_workout(db_path: Path) -> Workout:
    """Create a workout based on the Dan John Workout Generator.

    Args:
        db_path: The path to the database.

    Returns:
        A custom workout object built by the user.
    """
    data = read_database(db_path)
    bodyweight = data["loads"]["bodyweight"]
    units = data["loads"]["units"]
    workout_type, workout_params = _get_workout_params("wg")
    variation = _get_options(workout_params)
    if workout_params[variation].get("time"):
        time = workout_params[variation]["time"]
    else:
        time = IntPrompt.ask("Duration (mins)")
    exercises = []
    while True:
        name = _get_exercise(db_path)
        if not name:
            break
        console.print(name)
        load = IntPrompt.ask(f"  Load in {units}")
        sets = workout_params[variation]["sets"]
        reps = IntPrompt.ask("  (Average) Reps per set")
        load += _add_bodyweight_factor(bodyweight, name)
        exercises.append(Exercise(name=name, load=load, sets=sets, reps=reps))
    return Workout(
        workout_type=workout_type,
        variation=variation.title(),
        time=time,
        units=units,
        bodyweight=bodyweight,
        exercises=exercises,
    )


def create_rite_of_passage_workout(db_path: Path) -> Workout:
    """Create a Rite of Passage workout.

    Args:
        db_path: The Path to the database.

    Returns:
        A workout object.
    """
    data = read_database(db_path)
    bodyweight = data["loads"]["bodyweight"]
    units = data["loads"]["units"]
    workout_type, workout_params = _get_workout_params("rop")
    variation = _get_options(workout_params)
    time = IntPrompt.ask("Workout duration (mins)")
    exercises = []
    while True:
        name = _get_options(workout_params[variation]["exercises"])
        if name.lower() == "done":
            break
        load = IntPrompt.ask(f"  Load in {units}")
        if name in ["Kettlebell Swing", "Kettlebell Snatch"]:
            sets = IntPrompt.ask("  Number of sets")
            reps = IntPrompt.ask("  Number of reps")
        else:
            sets = IntPrompt.ask("  Number of ladders")
            rungs = IntPrompt.ask("  Number of rungs")
            reps = rungs * (rungs + 1) // 2
        if name.lower() == "kettlebell clean and press":
            reps *= 2
        exercise = Exercise(name=name, load=load, sets=sets, reps=reps)
        exercise.load += _add_bodyweight_factor(bodyweight, exercise.name)
        exercises.append(exercise)
    return Workout(
        workout_type=workout_type,
        variation=f"{variation:0>2}",
        time=time,
        units=units,
        bodyweight=bodyweight,
        exercises=exercises,
    )


def _add_bodyweight_factor(bodyweight, name):
    return int(BODYWEIGHT_FACTORS.get(name, 0) * bodyweight)


def add_exercises(db_path: Path, bodyweight, units) -> list[Exercise]:
    exercises = []
    while True:
        name = _get_exercise(db_path)
        if not name:
            break
        console.print(name)
        load = IntPrompt.ask(f"  Load in {units}")
        sets = IntPrompt.ask("  Number of sets")
        reps = IntPrompt.ask("  Reps per set")
        load += _add_bodyweight_factor(bodyweight, name)
        exercises.append(Exercise(name=name, load=load, sets=sets, reps=reps))
    return exercises


def create_custom_workout(db_path: Path) -> Workout:
    """Create a custom workout.

    Args:
        db_path: The path to the database.

    Returns:
        A custom workout object built by the user.
    """
    data = read_database(db_path)
    bodyweight = data["loads"]["bodyweight"]
    units = data["loads"]["units"]
    workout_type = Prompt.ask("Type of workout")
    if not workout_type:
        workout_type = "custom"
    variation = Prompt.ask("Variation")
    if not variation:
        variation = "Custom"
    time = IntPrompt.ask("Duration (mins)")
    exercises = add_exercises(db_path, bodyweight, units)

    return Workout(
        workout_type=workout_type,
        variation=variation.title(),
        time=time,
        units=units,
        bodyweight=bodyweight,
        exercises=exercises,
    )


def create_abf_barbell_workout(db_path: Path) -> Workout:
    """Create an Amor Building Formula Barbell workout.

    Args:
        db_path: The path to the database.

    Returns:
        A workout object built by the user.
    """
    data = read_database(db_path)
    bodyweight = data["loads"]["bodyweight"]
    units = data["loads"]["units"]
    workout_type, workout_param = _get_workout_params("abfb")
    variation = _get_options(workout_param["variations"])
    time = IntPrompt.ask("Duration (mins)")
    exercises = []
    if variation == "Program Three":
        squat_reps = IntPrompt.ask("Reps per set for front squat")
    for exercise in workout_param["exercises"]:
        print()
        print(exercise)
        if exercise == "Curl":
            sets_left = IntPrompt.ask(f"Number of total sets in the {exercise}")
        else:
            sets_left = workout_param["exercises"][exercise]["sets"]
        while sets_left > 0:
            load = IntPrompt.ask("  Load in lbs")
            sets = IntPrompt.ask("  Number of sets at this load")
            reps = workout_param["exercises"][exercise]["reps"]
            sets_left -= sets
            load = convert_pounds_to_kilos(load)
            exercises.append(Exercise(name=exercise, load=load, sets=sets, reps=reps))
            if variation == "Program Three" and exercise == "Clean and Press":
                exercises.append(
                    Exercise(name="Front Squat", load=load, sets=sets, reps=squat_reps)
                )
    if Confirm.ask("Did you do any auxiliary exercises"):
        auxiliary_exercises = add_exercises(db_path, bodyweight, units)
        exercises.extend(auxiliary_exercises)
    print()
    return Workout(
        workout_type=workout_type,
        variation=variation.title(),
        time=time,
        units=units,
        bodyweight=bodyweight,
        exercises=exercises,
    )


def create_abf_workout(db_path: Path) -> Workout:
    """Create an armor building formula Workout.

    Args:
        db_path: The Path object to the database.

    Returns:
        A Workout object created by the user.
    """
    data = read_database(db_path)
    bodyweight = data["loads"]["bodyweight"]
    units = data["loads"]["units"]
    workout_type, workout_params = _get_workout_params("abf")
    variation = _get_options(workout_params)
    time = IntPrompt.ask("How long was your workout (mins)")
    load = IntPrompt.ask(f"What weight did you use ({units})")
    exercises = []
    if variation in ("Armor Building Complex", "Intro"):
        sets = IntPrompt.ask("How many rounds of ABC did you complete?")
        for exercise, reps in workout_params["Armor Building Complex"]:
            exercises.append(
                Exercise(
                    name=exercise,
                    load=load,
                    sets=sets,
                    reps=reps,
                )
            )
    if variation in ("Double Kettlebell Press", "Intro"):
        for exercise, reps in workout_params["Double Kettlebell Press"]:
            sets = IntPrompt.ask(f"How many sets of {reps} reps did you complete")
            exercises.append(
                Exercise(
                    name=exercise,
                    load=load,
                    sets=sets,
                    reps=reps,
                )
            )
    return Workout(
        workout_type=workout_type,
        variation=variation,
        time=time,
        units=units,
        bodyweight=bodyweight,
        exercises=exercises,
    )


def create_template(db_path: Path) -> None:
    """Create a template for a program."""
    data = read_database(db_path)
    bodyweight = data["loads"]["bodyweight"]
    units = data["loads"]["units"]
    workout_type = Prompt.ask("Type of workout")
    if workout_type not in data["templates"]:
        data["templates"][workout_type] = {}
    variation = Prompt.ask("Variation")
    exercises = add_exercises(db_path, bodyweight, units)
    data["templates"][workout_type][variation] = [e.dict() for e in exercises]
    write_database(db_path, data)


def _get_options(options: dict | list) -> str:
    """Select options for a given workout parameter.
    Args:
        workout_param: options for a given workout parameter.
    Returns:
        A string consisting of the workout parameter choosen by the user.
    """
    if isinstance(options, dict):
        options = list(options.keys())
    for i, option in enumerate(options, 1):
        console.print(f"    [{i}] {option.title()}")
    while True:
        try:
            selection = IntPrompt.ask("Choose your option")
            return options[selection - 1]
        except (IndexError, TypeError):
            console.print(":warning: Not a valid option.", style=WARNING)
            console.print(
                "Enter a number between 1 and {max(options) + 1}.", style=SUGGESTION
            )
            continue


def _get_units() -> str:
    """A helper function to get the units.

    Returns:
        A string, either 'lbs' or 'kg'.
    """
    while True:
        units = Prompt.ask("[P]ounds or [K]ilograms").lower()
        if units.startswith("p"):
            units = "lbs"
        elif units.startswith("k"):
            units = "kg"
        else:
            console.print(":warning: Invalid option", style=WARNING)
            console.print("Please enter P or K", style=SUGGESTION)
            continue
        break
    return units


def _get_workout_params(workout_type: str) -> tuple[str, dict]:
    """Gets the workout parameters from the constants file.

    Args:
        workout_type: The workout type: ic, abc, btb, or pw.

    Returns:
        A tuple of the long name, and the dict of parameters.
    """
    match workout_type:
        case "ic":
            return "iron cardio", IRON_CARDIO_PARAMS
        case "abc":
            return "armor building complex", ABC_PARAMS
        case "btb":
            return "back to basics", BTB_PARAMS
        case "pw":
            return "perfect workout", PW_PARAMS
        case "giant":
            return "the giant", THE_GIANT_PARAMS
        case "wolf":
            return "the wolf", THE_WOLF_PARAMS
        case "es":
            data = read_database(KETTLEBELLS_DB)
            EASY_STRENGTH_PARAMS = data["templates"]["easy strength"]
            return "easy strength", EASY_STRENGTH_PARAMS
        case "wg":
            return "workout generator", WORKOUT_GENERATOR_PARAMS
        case "rop":
            return "rite of passage", ROP_PARAMS
        case "abfb":
            return "armor building formula barbell", ABFB_PARAMS
        case "abf":
            return "armor building formula", ABF_PARAMS
        case "dfw":
            return "dry fighting weight", DFW_PARAMS


def convert_pounds_to_kilos(load: int) -> int:
    """Convert a load in pounds to kilograms."""
    return round(load * POUNDS_TO_KILOS_RATE, 0)


def add_exercise_to_database(db_path: Path, exercise: str) -> None:
    """Add an exercise to the database."""
    data = read_database(db_path)
    movements = []
    while True:
        movement = _get_options(HUMAN_MOVEMENTS)
        if movement == "done":
            break
        else:
            movements.append(movement)
    data["exercises"][exercise] = movements
    write_database(db_path, data)


def _print_helper(to_print: list) -> None:
    """Print out various workout parameters formatted based on longest label.

    Args:
        to_print: A list of workout parameters.
    """
    longest_label = len(max(to_print, key=lambda x: len(x[0]))[0])
    for label, value in to_print:
        console.print(f"{label: >{longest_label}}: {value}")
    console.print()


def _get_exercise(db_path: Path) -> str | None:
    environ["FZF_DEFAULT_OPTS"] = FZF_DEFAULT_OPTS
    exercises = read_database(db_path)["exercises"]
    exercise = iterfzf(exercises, multi=False)
    match exercise:
        case "Other":
            exercise = Prompt.ask("Name of exercise").title()
            add_exercise_to_database(db_path, exercise)
            return exercise
        case "Done" | None:
            return None
        case _:
            return exercise



import sys
from datetime import datetime
from os import environ
from pathlib import Path

import typer
from rich.prompt import Confirm, Prompt
from trogon import Trogon
from typer.main import get_group
from typing_extensions import Annotated

from . import __version__
from .console import console
from .constants import (
    DATE_FORMAT,
    FZF_DEFAULT_OPTS,
    KETTLEBELLS_DB,
    KETTLEBELLS_HOME,
    SUGGESTION,
    WARNING,
)
from .database import (
    cache_workout,
    confirm_loads,
    initialize_database,
    update_database,
    read_database,
    save_workout,
    write_database,
)
from .stats import (
    filter_by_program,
    get_all_stats,
    get_exercises_by_movement,
    plot_workouts,
    print_calendar,
    retrieve_workout,
    top_ten_workouts,
    view_program,
)
from .workouts import (
    Workout,
    create_abf_barbell_workout,
    create_abf_workout,
    create_btb_workout,
    create_custom_workout,
    create_easy_strength_workout,
    create_ic_or_abc,
    create_perfect_workout,
    create_rite_of_passage_workout,
    create_set_based_workout,
    create_time_based_workout,
    create_template,
    create_workout_generator_workout,
    random_ic_or_abc,
    set_loads,
    set_program_loads,
)

cli = typer.Typer(add_completion=False)

environ["FZF_DEFAULT_OPTS"] = FZF_DEFAULT_OPTS


@cli.command()
def tui(ctx: typer.Context):
    Trogon(get_group(cli), click_context=ctx).run()


@cli.command()
def report_version(display: bool) -> None:
    """Print version and exit."""
    if display:
        console.print(f"{Path(sys.argv[0]).name} {__version__}")
        raise typer.Exit()


@cli.callback()
def global_options(
    ctx: typer.Context,
    version: bool = typer.Option(
        False,
        "--version",
        "-v",
        is_flag=True,
        is_eager=True,
        callback=report_version,
    ),
):
    """Create, save, and track progress of kettlebell workouts."""


@cli.command()
def init(
    ctx: typer.Context,
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        is_flag=True,
        is_eager=True,
        help="This will overwrite the existing database.",
    ),
) -> None:
    """Initializes the kettlebells database."""
    initialize_database(
        kettlebells_home=KETTLEBELLS_HOME,
        db_path=KETTLEBELLS_DB,
        force=force,
    )


@cli.command()
def update(
    ctx: typer.Context,
) -> None:
    """Update the database with new exercises."""
    update_database(KETTLEBELLS_DB)


@cli.command()
def setloads(
    ctx: typer.Context,
    program: bool = typer.Option(
        False,
        "--program",
        "-p",
        is_flag=True,
        is_eager=True,
        help="Add a load for a special program.",
    ),
) -> None:
    """Set units and loads for workouts."""
    data = read_database(KETTLEBELLS_DB)
    if program:
        data["loads"] = set_program_loads(data["loads"])
    else:
        loads = set_loads()
        data["loads"] = loads
    write_database(KETTLEBELLS_DB, data)


@cli.command()
def random(
    ctx: typer.Context,
    workout_type: Annotated[
        str,
        typer.Option("--workout-type", "-w", help="Possible workout types: ic or abc."),
    ],
) -> None:
    """Create a random iron cardio or armor building complex workout."""
    confirm_loads(KETTLEBELLS_DB)
    workout = random_ic_or_abc(KETTLEBELLS_DB, workout_type)
    cache_workout(KETTLEBELLS_DB, workout)
    workout.display_workout()


@cli.command()
def template(
    ctx: typer.Context,
) -> None:
    """Create a template for a program."""
    confirm_loads(KETTLEBELLS_DB)
    create_template(KETTLEBELLS_DB)


@cli.command()
def save(
    ctx: typer.Context,
    workout_type: Annotated[
        str,
        typer.Option(
            "--workout-type",
            "-w",
            help="Possible workout-types: ic, abc, btb, dfw, es, pw, wolf, wg, rop, or custom.",
        ),
    ] = "",
) -> None:
    """Save a kettlebell workout.

    If no argument is passed, kettlebells will attempt to use the most recently generated workout.
    """
    confirm_loads(KETTLEBELLS_DB)
    match workout_type:
        case "ic" | "abc":
            workout = create_ic_or_abc(KETTLEBELLS_DB, workout_type)
        case "btb":
            workout = create_btb_workout(KETTLEBELLS_DB)
        case "custom":
            workout = create_custom_workout(KETTLEBELLS_DB)
        case "pw":
            workout = create_perfect_workout(KETTLEBELLS_DB)
        case "giant" | "dfw":
            workout = create_time_based_workout(KETTLEBELLS_DB, workout_type)
        case "wolf":
            workout = create_set_based_workout(KETTLEBELLS_DB, workout_type)
        case "es":
            workout = create_easy_strength_workout(KETTLEBELLS_DB, workout_type)
        case "wg":
            workout = create_workout_generator_workout(KETTLEBELLS_DB)
        case "rop":
            workout = create_rite_of_passage_workout(KETTLEBELLS_DB)
        case "abfb":
            workout = create_abf_barbell_workout(KETTLEBELLS_DB)
        case "abf":
            workout = create_abf_workout(KETTLEBELLS_DB)
        case "":
            data = read_database(KETTLEBELLS_DB)
            workout = Workout(**data["cached_workouts"][-1])
            console.print("Last workout generated:\n")
        case _:
            console.print(
                f"'{workout_type.strip()}' is not an option.",
                style=WARNING,
            )
            console.print(
                "Try running [underline]kettlebells save --help[/underline]",
                style=SUGGESTION,
            )
            return

    if workout is None:
        return
    workout.display_workout()
    if Confirm.ask("Save this workout?"):
        workout_date = _get_date()
        save_workout(KETTLEBELLS_DB, workout_date, workout)
        print()
        workout.display_workout_stats()
    else:
        console.print("Workout not saved.")


@cli.command()
def view(
    ctx: typer.Context,
    preview: Annotated[
        bool,
        typer.Option(
            "--preview",
            "-p",
            help="Display a preview of workouts in database.",
            is_flag=True,
        ),
    ] = False,
    program: Annotated[
        bool,
        typer.Option(
            "--program",
            "-P",
            help="Filter out workouts of a certain program.",
            is_flag=True,
        ),
    ] = False,
    display_workout: Annotated[
        bool,
        typer.Option(
            "--display_workout",
            "-d",
            help="Display individual workouts in a program.",
            is_flag=True,
        ),
    ] = False,
) -> None:
    """Display stats of a given workout or program."""
    data = read_database(KETTLEBELLS_DB)
    if program:
        data, program_name = filter_by_program(data)
        table = view_program(data, program_name, display_workout)
        console.print(table)
        get_all_stats(data, program_name)
        return
    try:
        date, workout = retrieve_workout(data, preview)
        console.print(
            f"\nDate: [green]{datetime.strptime(date, DATE_FORMAT):%b %d, %Y}\n"
        )
        workout.display_workout()
        print()
        workout.display_workout_stats()
    except TypeError:
        return


@cli.command()
def last(ctx: typer.Context) -> None:
    """Display stats from most recent workout in database."""
    data = read_database(KETTLEBELLS_DB)
    last_workout = data["saved_workouts"][-1]
    workout_date = last_workout["date"]
    workout = Workout(**last_workout["workout"])
    console.print(
        f"\nDate: [green]{datetime.strptime(workout_date, DATE_FORMAT):%b %d, %Y}\n"
    )
    workout.display_workout()
    print()
    workout.display_workout_stats()


@cli.command()
def stats(
    ctx: typer.Context,
    plot: Annotated[
        str,
        typer.Option(
            "--plot",
            "-p",
            help="Possible plots: line, event, bar.",
        ),
    ] = "",
    year: Annotated[
        int | None,
        typer.Option(
            "--year",
            "-y",
            help="Year to display calendar and event plot for.",
        ),
    ] = None,
    sort: Annotated[
        str,
        typer.Option(
            "--sort",
            "-S",
            help="Sort the best table. Possible arguments: weight-moved, reps, weight-density, rep-density.",
        ),
    ] = "weight-moved",
    start: Annotated[
        str,
        typer.Option(
            "--start",
            "-s",
            help="Start date (YYYY-MM-DD).",
        ),
    ] = "",
    end: Annotated[
        str,
        typer.Option(
            "--end",
            "-e",
            help="End date (YYYY-MM-DD).",
        ),
    ] = "",
    calendar: Annotated[
        bool,
        typer.Option(
            "--calendar",
            "-c",
            help="Highlight days workouts were completed in a calendar.",
            is_flag=True,
        ),
    ] = False,
    by_movement: Annotated[
        bool,
        typer.Option(
            "--by-movement",
            "-M",
            help="Count the reps by movement.",
            is_flag=True,
        ),
    ] = False,
    median: Annotated[
        bool,
        typer.Option(
            "--median",
            "-m",
            help="A horizontal line at the median weight per workout. Use in conjunction with --plot.",
            is_flag=True,
        ),
    ] = False,
    average: Annotated[
        bool,
        typer.Option(
            "--average",
            "-a",
            help="A horizontal line at the mean weight per workout. Use in conjunction with --plot.",
            is_flag=True,
        ),
    ] = False,
    best: Annotated[
        bool,
        typer.Option(
            "--best",
            "-b",
            help="Show the top ten workouts in a table.",
            is_flag=True,
        ),
    ] = False,
) -> None:
    """Display stats from all workouts in database."""
    data = read_database(KETTLEBELLS_DB)
    if calendar and not year:
        year = datetime.now().year
    if year:
        dates, weight_per_workout = get_all_stats(
            data,
            start=f"{year}-01-01",
            end=f"{year}-12-31",
            year=year,
        )
    else:
        dates, weight_per_workout = get_all_stats(data, start=start, end=end)
    if plot:
        plot_workouts(dates, weight_per_workout, plot, median, average, year)
    if calendar:
        print_calendar(data, year)
    if best:
        console.print(top_ten_workouts(data, sort))
    if by_movement and year:
        start = f"{year}-01-01"
        end = f"{year}-12-31"
        console.print(get_exercises_by_movement(data, start, end))
    if by_movement and not year:
        console.print(get_exercises_by_movement(data, start, end))


def _get_date() -> str:
    """A helper function to get the date of a workout.

    Returns:
        A str of the date formatted as YYYY-MM-DD.
    """
    while True:
        workout_date = Prompt.ask(
            "Enter the date of the workout (YYYY-MM-DD), or press enter for today"
        )
        if not workout_date:
            workout_date = datetime.now().strftime(DATE_FORMAT)
        try:
            datetime.strptime(workout_date, DATE_FORMAT)
            break
        except ValueError:
            console.print(f":warning: {workout_date} not a valid date.", style=WARNING)
            continue
    return workout_date


if __name__ == "__main__":
    cli()



import json
import sys
from collections import deque
from datetime import datetime
from pathlib import Path

from .console import console
from .constants import DATE_FORMAT, EXERCISES, SUGGESTION, WARNING


def initialize_database(kettlebells_home: Path, db_path: Path, force: bool) -> None:
    """Creates the home directory and the JSON database.

    Args:
        kettlebells_home: The home directory for the kettlebells database.
        db_path: The Path to the database.
        force: If True, overwrites the existing database with a blank one.

    """
    if kettlebells_home.is_dir() and force:
        pass
    elif db_path.is_file():
        console.print(":warning: Database base already exits.", style=WARNING)
        console.print(
            "Run [underline]kettlebells init --force[/underline] to overwrite database.",
            style=SUGGESTION,
        )
        return
    try:
        kettlebells_home.mkdir()
    except FileExistsError:
        pass
    data: dict = {
        "loads": dict(),
        "templates": dict(),
        "exercises": EXERCISES,
        "saved_workouts": [],
        "cached_workouts": [],
    }
    write_database(db_path, data)


def update_database(db_path: Path) -> None:
    """Update the database with new exercises.

    Args:
        db_path: The Path to the database.
        data: The new version of the database to write.

    """
    data = read_database(db_path)
    exercises = data.get("exercises")
    if exercises:
        data["exercises"] = exercises | EXERCISES
    else:
        data["exercises"] = EXERCISES
    write_database(db_path, data)


def read_database(db_path: Path) -> dict:
    """Read from the data base.

    Args:
        db_path: The Path to the database.

    Returns:
        A json object of the data.

    """
    if not db_path.is_file():
        console.print(":warning: Could not find kettlebells database.", style=WARNING)
        console.print(
            "Try running [underline]kettlebells init[/underline] first.",
            style=SUGGESTION,
        )
        sys.exit()
    with open(db_path) as db:
        return json.load(db)


def write_database(db_path: Path, data: dict) -> None:
    """Write data to database.
    :param db_path: The Path to the database.
    :param data: The new version of the database to write.
    :returns: None
    """
    with open(db_path, "w") as db:
        json.dump(data, db, indent=4)


def confirm_loads(db_path: Path) -> None:
    """Checks if the loads have been set in the database.

    Args:
        db_path: The Path to the database.

    """
    data = read_database(db_path)
    if not data["loads"]:
        console.print("[red]:warning: Could not find loads in database.", style=WARNING)
        console.print(
            "Try running [underline]iron-cardio setloads[/underline] first.",
            style=SUGGESTION,
        )
        sys.exit()


def cache_workout(db_path: Path, workout) -> None:
    """Cache last 10 generated workout.

    Args:
        db_path: The Path to the database.
        workout: Workout object to be stored in the cache.

    """
    data = read_database(db_path)
    cache = deque(data["cached_workouts"], maxlen=1)
    cache.append(workout.model_dump())
    data["cached_workouts"] = list(cache)
    write_database(db_path, data)


def save_workout(db_path: Path, workout_date: str, workout) -> None:
    """Save a workout in the database.

    Args:
        db_path: The Path to the database.
        workout_date: The date of the workout.
        workout: Workout object to be stored in the database.

    """
    data = read_database(db_path)
    data["saved_workouts"].append(
        {"date": workout_date, "workout": workout.model_dump()}
    )
    data["saved_workouts"] = sorted(
        data["saved_workouts"], key=lambda x: datetime.strptime(x["date"], DATE_FORMAT)
    )
    write_database(db_path, data)



import json
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory

import pytest

from .test_constants import TEST_DATA


@pytest.fixture(scope="function")
def database_home():
    with TemporaryDirectory() as db_dir:
        db_home = Path(db_dir)
        db_path = db_home / "test_db.json"
        return db_path


@pytest.fixture(scope="function")
def database():
    database = NamedTemporaryFile(delete=False)
    with open(database.name, "w") as db:
        json.dump(TEST_DATA, db)
    return database


@pytest.fixture()
def prompt_mock(mocker):
    yield mocker.patch("kettlebells.workouts.Prompt.ask")


@pytest.fixture()
def int_mock(mocker):
    yield mocker.patch("kettlebells.workouts.IntPrompt.ask")


@pytest.fixture()
def confirm_mock(mocker):
    yield mocker.patch("kettlebells.workouts.Confirm")


@pytest.fixture()
def units_mock(mocker):
    yield mocker.patch("kettlebells.workouts._get_units")


@pytest.fixture()
def options_mock(mocker):
    yield mocker.patch("kettlebells.workouts._get_options")



from pathlib import Path
from unittest import mock

import pytest
from kettlebells.database import read_database
from kettlebells.stats import (
    _get_dates,
    filter_by_program,
    get_all_stats,
    plot_workouts,
    retrieve_workout,
    top_ten_workouts,
    view_program,
)
from kettlebells.workouts import Workout


def test_get_all_time_stats(database, capfd):
    """Test that get_all_time_stats returns the dates and stats from the test database."""
    expected = (["2023-09-12", "2023-09-14"], [3456, 3480])
    data = read_database(Path(database.name))
    actual = get_all_stats(data)
    output = capfd.readouterr()[0]
    expected_output = """\nAll Time Stats
==============
     Total Workouts: 2
         Total Time: 0 days 00 hours 40 mins
 Total Weight Moved: 6,936 kg
         Total Reps: 159
Mean Weight Density: 173.4 kg/min
   Mean Rep Density: 4.0 reps/min"""
    assert actual == expected
    assert expected_output in output


def test_top_ten_workouts(database):
    """Test the table object created by top_ten_workouts."""
    data = read_database(Path(database.name))
    table = top_ten_workouts(data, "weight-moved")
    assert table.title == "Top Ten Workouts by Weight Moved"
    assert len(table.rows) == 2
    assert len(table.columns) == 8


@mock.patch("kettlebells.stats.iterfzf")
def test_filter_by_program(fzf_mock, database):
    """Test the table object created by filter_by_program."""
    data = read_database(Path(database.name))
    data["saved_workouts"].pop(0)
    fzf_mock.return_value = "Dry Fighting Weight"

    filtered_data, program = filter_by_program(data)
    assert program == "Dry Fighting Weight"
    assert filtered_data == data


def test_view_program(database):
    """Test the table object created by filter_by_program."""
    data = read_database(Path(database.name))
    data["saved_workouts"].pop(0)
    table = view_program(data, "Dry Fighting Weight", False)
    assert table.title == "Dry Fighting Weight"
    assert len(table.rows) == 1
    assert len(table.columns) == 9


@mock.patch("kettlebells.stats.iterfzf")
def test_retrieve_workout(fzf_mock, database):
    """Test the table object created by filter_by_program."""
    data = read_database(Path(database.name))
    expected = Workout(**data["saved_workouts"][1]["workout"])
    fzf_mock.return_value = "2023-09-14"
    date, actual = retrieve_workout(data, False)
    assert actual == expected
    assert date == "2023-09-14"


@pytest.mark.parametrize(
    "title, x_label, plot_size, plot_type, median, average",
    [
        ("Workouts Across the Year", "Months", (130, 20), "event", False, False),
        ("Weight Moved Per Workout", "Date", (90, 30), "line", True, False),
        ("Weight Moved by Month", "Weight Moved", (90, 30), "bar", False, True),
    ],
)
@mock.patch("kettlebells.stats.plt")
def test_event_plot(plt_mock, title, x_label, plot_size, plot_type, median, average):
    plot_workouts(["2023-02-01"], [2300], plot_type, median, average)
    plt_mock.title.assert_called_once_with(title)
    plt_mock.xlabel.assert_called_once_with(x_label)
    plt_mock.plotsize.assert_called_once_with(*plot_size)
    if median:
        plt_mock.hline.assert_called_once()
    if average:
        plt_mock.vline.assert_called_once()


def test_dates_in_cal(database, capfd):
    data = read_database(Path(database.name))
    actual = _get_dates(data)
    expected = [(12, 9, 2023), (14, 9, 2023)]
    assert actual == expected



import json
from pathlib import Path
from unittest import mock

import pytest

import kettlebells.database as db
from kettlebells.constants import EXERCISES

from .test_constants import TEST_DATA, TEST_IC_WORKOUT, TEST_NEW_EXERCISES


def test_initialize_database(database_home):
    expected = {
        "loads": dict(),
        "templates": dict(),
        "exercises": EXERCISES,
        "saved_workouts": [],
        "cached_workouts": [],
    }
    db.initialize_database(database_home.parents[0], database_home, False)
    assert database_home.is_file()
    assert json.load(open(database_home)) == expected


def test_initialize_database_already_existes(database, capfd):
    db.initialize_database(Path(database.name).parents[0], Path(database.name), False)
    output = capfd.readouterr()[0]
    assert "Database base already exits." in output


def test_initialize_database_force(database):
    expected = {
        "loads": dict(),
        "templates": dict(),
        "exercises": EXERCISES,
        "saved_workouts": [],
        "cached_workouts": [],
    }
    db.initialize_database(Path(database.name).parents[0], Path(database.name), True)
    assert Path(database.name).is_file()
    assert json.load(open(database.name)) == expected


def test_read_no_database(database_home, capfd):
    with pytest.raises(SystemExit):
        db.read_database(database_home)
    output = capfd.readouterr()[0]
    assert "Could not find kettlebells database." in output


def test_confirm_loads(database_home, capfd):
    db.initialize_database(database_home.parents[0], database_home, False)
    with pytest.raises(SystemExit):
        db.confirm_loads(database_home)
    output = capfd.readouterr()[0]
    assert "Could not find loads in database." in output


def test_write_database(database):
    db.write_database(database.name, TEST_DATA)
    actual = json.load(open(database.name))
    assert actual == TEST_DATA


def test_read_database(database):
    data = db.read_database(Path(database.name))
    assert data == TEST_DATA


def test_save_workout(database):
    db.save_workout(Path(database.name), "2023-09-14", TEST_IC_WORKOUT)
    data = json.load(open(database.name))
    assert data["saved_workouts"][-1] == {
        "date": "2023-09-14",
        "workout": {
            "bodyweight": 90,
            "units": "kg",
            "variation": "Double Classic + Pullup",
            "time": 30,
            "exercises": [
                {
                    "name": "Double Kettlebell Clean",
                    "load": 28,
                    "sets": 20,
                    "reps": 1,
                },
                {
                    "name": "Double Kettlebell Press",
                    "load": 28,
                    "sets": 20,
                    "reps": 1,
                },
                {
                    "name": "Double Kettlebell Front Squat",
                    "load": 28,
                    "sets": 20,
                    "reps": 1,
                },
                {
                    "name": "Pullup",
                    "load": 86,
                    "sets": 20,
                    "reps": 1,
                },
                {
                    "name": "Kettlebell Swing",
                    "load": 28,
                    "sets": 1,
                    "reps": 60,
                },
            ],
            "workout_type": "iron cardio",
        },
    }


def test_cache_workout(database):
    db.cache_workout(Path(database.name), TEST_IC_WORKOUT)
    data = json.load(open(database.name))
    assert len(data["cached_workouts"]) == 1
    assert data["cached_workouts"][-1] == TEST_IC_WORKOUT.model_dump()


@mock.patch("kettlebells.database.EXERCISES", TEST_NEW_EXERCISES)
def test_update_database(database):
    db.update_database(Path(database.name))
    data = json.load(open(database.name))
    expected = {
        "Kettlebell Press": ["push"],
        "Goblet Squat": ["squat"],
    } | TEST_NEW_EXERCISES
    assert data["exercises"] == expected



import json
from pathlib import Path
from unittest import mock

import pytest

from kettlebells.constants import IRON_CARDIO_PARAMS
from kettlebells.workouts import (
    Workout,
    _get_options,
    _get_units,
    add_exercise_to_database,
    create_abf_barbell_workout,
    create_btb_workout,
    create_custom_workout,
    random_ic_or_abc,
    create_ic_or_abc,
    create_perfect_workout,
    create_set_based_workout,
    create_time_based_workout,
    set_loads,
    set_program_loads,
    create_easy_strength_workout,
    create_workout_generator_workout,
    create_rite_of_passage_workout,
    create_abf_workout,
)

from .test_constants import (
    TEST_ABFB,
    TEST_ABF_PRESS_WORKOUT,
    TEST_BTB_WORKOUT,
    TEST_CUSTOM_WORKOUT,
    TEST_DOUBLE_TRAVELING_2S_WORKOUT,
    TEST_GIANT_WORKOUT,
    TEST_IC_WORKOUT,
    TEST_PERFECT_WORKOUT,
    TEST_SINGLE_TRAVELING_2S_WORKOUT,
    TEST_WOLF_WORKOUT,
    TEST_WORKOUT_NO_SWINGS,
    TEST_WORKOUT_SINGLE_BELL_PULLUPS,
    TEST_EASY_STRENGTH_WORKOUT,
    TEST_WORKOUT_GENERATOR_WORKOUT,
    TEST_ROP_WORKOUT,
    TEST_ABF_ABC_WORKOUT,
)

POSSIBLE_SWINGS = [50, 60, 70, 80, 90, 100, 110, 120, 130, 140, 150]


def test_create_random_workout(database):
    """Test when a workout is created, the parameters are appropriate based on the
    database and within the ranges defined in the constants module.
    """
    actual = random_ic_or_abc(Path(database.name), "ic")
    assert isinstance(actual, Workout)
    assert (
        actual.variation in IRON_CARDIO_PARAMS["doublebell variations"].keys()
        or actual.variation in IRON_CARDIO_PARAMS["singlebell variations"].keys()
    )
    assert actual.time in IRON_CARDIO_PARAMS["times"].keys()


@pytest.mark.parametrize("workout", [TEST_IC_WORKOUT, TEST_BTB_WORKOUT])
def test_display_workout(workout, capfd):
    """Test a workout is displayed correctly in the console."""
    workout.display_workout()
    output = capfd.readouterr()[0]
    assert workout.workout_type.upper() in output
    assert "=" * len(workout.workout_type) in output
    assert "Variation: " in output
    assert "Time: " in output
    assert "Load: " in output


def test_display_custom_workout(capfd):
    """Test a custom workout is displayed correctly in the console."""
    TEST_CUSTOM_WORKOUT.display_workout()
    output = capfd.readouterr()[0]
    assert TEST_CUSTOM_WORKOUT.workout_type.upper() in output
    assert "=" * len(TEST_CUSTOM_WORKOUT.workout_type) in output
    assert "Variation: " in output
    assert "Time: " in output
    assert "Exercises" in output


def test_display_workout_no_swings(capfd):
    """Test a workout is displayed correctly in the console with no swings."""
    TEST_WORKOUT_NO_SWINGS.display_workout()
    output = capfd.readouterr()[0]
    assert TEST_WORKOUT_NO_SWINGS.workout_type.upper() in output
    assert "=" * len(TEST_WORKOUT_NO_SWINGS.workout_type) in output
    assert "Variation: " in output
    assert "Time: " in output
    assert "Load: " in output
    assert "Swing: " not in output


@pytest.mark.parametrize(
    "workout, stats",
    [
        (
            TEST_IC_WORKOUT,
            {
                "weight moved": 6760,
                "reps": 140,
                "weight density": 225.3,
                "rep density": 4.7,
            },
        ),
        (
            TEST_WORKOUT_NO_SWINGS,
            {
                "weight moved": 1536,
                "reps": 64,
                "weight density": 76.8,
                "rep density": 3.2,
            },
        ),
    ],
)
def test_calc_workout_stats(workout, stats):
    """Test workout stats are calculated correctly."""
    actual = workout.calc_workout_stats()
    expected = stats
    assert actual == expected


def test_display_workout_stats_single_bell_pullups(capfd):
    """Test workout stats are displayed correctly."""
    TEST_WORKOUT_SINGLE_BELL_PULLUPS.display_workout_stats()
    output = capfd.readouterr()[0]
    expected = """Workout Stats
=============
  Weight Moved: 1,050 kg
    Total Reps: 35
Weight Density: 105.0 kg/min
   Rep Density: 3.5 reps/min
"""
    assert expected in output


def test_set_loads(units_mock, confirm_mock, int_mock):
    """Test that setting the loads in the database works."""
    expected = {
        "units": "pounds",
        "bodyweight": 190,
        "light load": 1,
        "medium load": 2,
        "heavy load": 3,
    }
    int_mock.side_effect = [190, 1, 2, 3]  # [body_weight, light, med, heavy]
    confirm_mock.side_effect = ["y"]
    units_mock.side_effect = ["pounds"]
    actual = set_loads()
    assert actual == expected


@pytest.mark.parametrize("response, units", [("p", "lbs"), ("k", "kg")])
def test_get_units_good_input(response, units, prompt_mock):
    """Test that units are set correctly."""
    expected = units
    prompt_mock.side_effect = [response]
    actual = _get_units()
    assert actual == expected


@pytest.mark.parametrize(
    "workout_param, response, option",
    [
        (IRON_CARDIO_PARAMS["bells"], 1, "Single Bell"),
        (IRON_CARDIO_PARAMS["singlebell variations"], 3, "Classic + Snatch"),
    ],
)
def test_get_options(workout_param, response, option, int_mock):
    """Test the options for workout parameters are valid."""
    expected = option
    int_mock.side_effect = [response]
    actual = _get_options(workout_param)
    assert actual == expected


def test_get_options_bad_input(int_mock):
    """Test bad index results in infinite loop."""
    try:
        int_mock.side_effect = [10, 20, "k"]
        _get_options(IRON_CARDIO_PARAMS["bells"])
    except StopIteration:
        pass


@pytest.mark.parametrize(
    "test_workout, bells, variation, confirm, int_responses",
    [
        # int responses are time, load, sets, swings count, swing load
        (
            TEST_IC_WORKOUT,
            "Double Bells",
            "Double Classic + Pullup",
            True,
            [30, 28, 20, 60, 28],
        ),
        (
            TEST_DOUBLE_TRAVELING_2S_WORKOUT,
            "Double Bells",
            "Double Traveling 2s",
            True,
            [29, 28, 12, 50, 28],
        ),
        (
            TEST_SINGLE_TRAVELING_2S_WORKOUT,
            "Single Bell",
            "Traveling 2s",
            True,
            [10, 20, 10, 100, 20],
        ),
    ],
)
def test_custom_ic_workout(
    test_workout,
    bells,
    variation,
    confirm,
    int_responses,
    options_mock,
    units_mock,
    confirm_mock,
    int_mock,
    database,
):
    """Test creating a custom iron cardio or abc workout works as intended."""
    expected = test_workout
    options_mock.side_effect = [bells, variation]
    int_mock.side_effect = int_responses
    units_mock.side_effect = ["kg"]
    confirm_mock.return_value = "y"
    actual = create_ic_or_abc(Path(database.name), "ic")
    assert actual == expected


# def test_easy_strength_workout(
#     options_mock,
#     units_mock,
#     confirm_mock,
#     int_mock,
#     database,
# ):
#     """Test creating an easy strength workout."""
#     expected = TEST_EASY_STRENGTH_WORKOUT
#     options_mock.side_effect = ["regular"]
#     int_mock.side_effect = [30, 24, 20, 24, 20, 24, 20, 24, 20, 24, 20, 24]
#     units_mock.side_effect = ["kg"]
#     actual = create_easy_strength_workout(Path(database.name), "es")
#     assert isinstance(actual, Workout)
#     assert actual == expected


@mock.patch("kettlebells.workouts.iterfzf")
def test_workout_generator_workout(
    fzf_mock,
    options_mock,
    units_mock,
    confirm_mock,
    int_mock,
    database,
):
    """Test creating a workout generator workout."""
    expected = TEST_WORKOUT_GENERATOR_WORKOUT
    int_mock.side_effect = [40, 100, 10, 80, 10]
    options_mock.side_effect = ["3 X 10"]
    fzf_mock.side_effect = ["Deadlift", "Bench Press", "Done"]
    units_mock.side_effect = ["kg"]
    actual = create_workout_generator_workout(Path(database.name))
    assert actual == expected


def test_rop_workout(
    options_mock,
    units_mock,
    int_mock,
    database,
):
    """Test creating a rite of passage workout."""
    expected = TEST_ROP_WORKOUT
    options_mock.side_effect = [
        "Medium",
        "Kettlebell Clean and Press",
        "Pullup",
        "Kettlebell Swing",
        "Done",
    ]
    int_mock.side_effect = [45, 28, 5, 3, 0, 5, 3, 28, 5, 10]
    units_mock.side_effect = ["kg"]
    actual = create_rite_of_passage_workout(Path(database.name))
    assert actual == expected


@pytest.mark.parametrize(
    "test_workout, variation, int_responses",
    [
        (
            TEST_ABF_ABC_WORKOUT,
            "Armor Building Complex",
            [30, 20, 30],
        ),
        (
            TEST_ABF_PRESS_WORKOUT,
            "Double Kettlebell Press",
            [30, 20, 5, 5, 5, 5],
        ),
    ],
)
def test_abf_workout(
    test_workout,
    variation,
    int_responses,
    options_mock,
    units_mock,
    int_mock,
    database,
):
    """Test creating a armor building formula workout."""
    expected = test_workout
    options_mock.side_effect = [variation]
    int_mock.side_effect = int_responses
    units_mock.side_effect = ["kg"]
    actual = create_abf_workout(Path(database.name))
    assert actual == expected


@mock.patch("kettlebells.workouts.iterfzf")
def test_abfb(
    fzf_mock,
    options_mock,
    units_mock,
    confirm_mock,
    int_mock,
    database,
):
    """Test creating a armor building formula barbell workout."""
    expected = TEST_ABFB
    options_mock.side_effect = ["Program Three"]
    int_mock.side_effect = [
        32,  # time
        1,  # squat reps
        45,  # load
        1,  # sets at ^ load
        65,  # load
        1,  # sets at ^ load
        95,  # load
        3,  # sets at ^ load
        5,  # sets of curls
        45,  # load
        2,  # sets at ^ load
        55,  # load
        3,  # sets at ^ load
    ]
    units_mock.side_effect = ["kg"]
    confirm_mock.return_value = "n"
    fzf_mock.return_value = None
    actual = create_abf_barbell_workout(Path(database.name))
    assert isinstance(actual, Workout)
    assert actual == expected


def test_btb_workout(
    options_mock,
    units_mock,
    confirm_mock,
    int_mock,
    database,
):
    """Test creating a back to basics workout works as intended."""
    expected = TEST_BTB_WORKOUT
    options_mock.side_effect = ["2 C&P Ladders + Snatch"]
    int_mock.side_effect = [30, 24, 20]  # time, load, sets
    units_mock.return_value = "kg"
    actual = create_btb_workout(Path(database.name))
    assert isinstance(actual, Workout)
    assert actual == expected


def test_perfect_workout(
    options_mock,
    units_mock,
    int_mock,
    database,
):
    """Test creating a perfect."""
    expected = TEST_PERFECT_WORKOUT
    options_mock.side_effect = ["The Bull"]
    int_mock.side_effect = [10, 20, 24, 20, 20]  # time, press, BGBS, Row, GS
    units_mock.return_value = "kg"
    actual = create_perfect_workout(Path(database.name))
    assert isinstance(actual, Workout)
    assert actual == expected


@pytest.mark.parametrize(
    "workout_type, variation", [("custom", "custom"), (None, None)]
)
@mock.patch("kettlebells.workouts.iterfzf")
def test_create_custom_workout(
    fzf_mock,
    workout_type,
    variation,
    int_mock,
    prompt_mock,
    database,
):
    """Test creating a custom workout works as intended."""
    expected = TEST_CUSTOM_WORKOUT
    prompt_mock.side_effect = [workout_type, variation]
    fzf_mock.side_effect = ["Turkish Get-up", "TRX T", "Done"]
    # int_mocks = [time, load1, sets1, reps1, load2, sets2, reps3]
    int_mock.side_effect = [30, 24, 1, 6, 0, 3, 8]
    actual = create_custom_workout(Path(database.name))
    assert isinstance(actual, Workout)
    assert actual == expected


def test_set_progam_loads(int_mock, prompt_mock):
    """Test if setting program loads addes to loads dict."""
    prompt_mock.side_effect = ["the giant"]
    int_mock.side_effect = [24]
    loads = {
        "units": "kg",
        "bodyweight": 90,
        "light load": 20,
        "medium load": 24,
        "heavy load": 28,
    }
    expected = {
        "units": "kg",
        "bodyweight": 90,
        "light load": 20,
        "medium load": 24,
        "heavy load": 28,
        "the giant": 24,
    }
    actual = set_program_loads(loads)
    assert actual == expected


@mock.patch("kettlebells.workouts.iterfzf")
def test_create_giant_workout(
    iterfzf_mock, int_mock, prompt_mock, confirm_mock, database
):
    """Test creating a giant workout."""
    iterfzf_mock.return_value = None
    prompt_mock.side_effect = ["1", "1"]
    int_mock.side_effect = [1, 30, 10]
    confirm_mock.side_effect = "n"
    actual = create_time_based_workout(Path(database.name), "giant")
    assert actual == TEST_GIANT_WORKOUT


def test_create_wolf_workout(int_mock, prompt_mock, database):
    """Test creating a wolf workout."""
    prompt_mock.side_effect = ["1", "1"]
    int_mock.side_effect = [1, 12]
    actual = create_set_based_workout(Path(database.name), "wolf")
    assert actual == TEST_WOLF_WORKOUT


# def test_add_exercise_to_database(options_mock, database):
#     """Test adding an exercise to the database."""
#     options_mock.side_effect = ["squat", "Done"]
#     add_exercise_to_database(Path(database.name), "leg press")
#     db = json.load(open(database.name))
#     assert db["exercises"]["leg curl"] == ["lower body isolation"]



import calendar
import sys
from collections import defaultdict
from datetime import datetime
from statistics import mean, median

import pandas as pd
import plotext as plt
from iterfzf import iterfzf
from rich import box
from rich.align import Align
from rich.columns import Columns
from rich.table import Table
from rich.text import Text

from .console import console
from .constants import DATE_FORMAT, KETTLEBELLS_HOME, SUGGESTION, WARNING
from .workouts import Workout, _print_helper


def get_all_stats(
    data: dict,
    program: str | None = None,
    start: str | None = None,
    end: str | None = None,
    year: int | None = None,
) -> tuple[list[str], list[int]]:
    """Print stats from all workout in the database between start_date and end_date.

    Args:
        data: A dict of the data in the database.
        program: A str of the program to filter by.
        start: A str of the start date to filter by.
        end: A str of the end date to filter by.
        year: A int of the year to display.

    Returns:
        tuple: Lists of both dates and weight moved per workout.
    """
    units = data["loads"]["units"]
    dates = []
    stats = []
    workouts = []
    if start:
        start_date = datetime.strptime(start, DATE_FORMAT)
    if end:
        end_date = datetime.strptime(end, DATE_FORMAT)
    for workout_data in data["saved_workouts"]:
        date = workout_data["date"]
        date_obj = datetime.strptime(date, DATE_FORMAT)
        if start and date_obj < start_date:
            continue
        if end and date_obj > end_date:
            continue
        dates.append(date)
        workout = Workout(**workout_data["workout"])
        stats.append(workout.calc_workout_stats())
        workouts.append(workout)

    weight_per_workout = [stat["weight moved"] for stat in stats]
    total_weight_moved = sum(weight_per_workout)
    total_reps = sum(stat["reps"] for stat in stats)
    average_weight_density = mean(stat["weight density"] for stat in stats)
    average_rep_density = mean(stat["rep density"] for stat in stats)

    # Calculate total time
    days, remaining_mins = divmod(sum(workout.time for workout in workouts), 60 * 24)
    hours, mins = divmod(remaining_mins, 60)

    # Create title for stats
    if program:
        title = f"{program.title()} Stats"
    elif year:
        title = f"Stats from the year {year}"
    elif start and end:
        title = f"Stats from {start} to {end}"
    elif start:
        title = f"Stats from {start} to Present"
    elif end:
        title = f"Stats to {end}"
    else:
        title = "All Time Stats"

    console.print()
    console.print(title)
    console.print("=" * len(title), style="green")
    stats_to_print = [
        ("Total Workouts", f"{len(stats):,}"),
        ("Total Time", f"{days} days {hours:02} hours {mins:02} mins"),
        ("Total Weight Moved", f"{total_weight_moved:,} {units}"),
        ("Total Reps", f"{total_reps:,}"),
        ("Mean Weight Density", f"{average_weight_density:.1f} {units}/min"),
        ("Mean Rep Density", f"{average_rep_density:.1f} reps/min"),
    ]
    _print_helper(stats_to_print)
    return dates, weight_per_workout


def plot_workouts(
    dates: list[str],
    weight_per_workout: list[int],
    plot_type: str,
    show_median: bool,
    show_average: bool,
    year: int | None = None,
) -> None:
    """Plot weight moved per workout.

    Args:
        dates: A list of the dates stored as a string.
        weight_per_workout: A list of the weight moved per workout.
    """
    match plot_type:
        case "event":
            x_axis = []
            for date in dates:
                year, month, day = date.split("-")
                x_axis.append("-".join((year[-2:], month, day)))
            plt.date_form("y-m-d")
            plt.plotsize(130, 20)
            plt.title("Workouts Across the Year")
            plt.xlabel("Months")
            if year and len(year) == 4:
                year = year[-2:]
            if not year:
                year = str(datetime.today().year)[-2:]
            plt.xlim(f"{year}-01-01", f"{year}-12-31")
            plt.event_plot(x_axis, marker="hd")
            ticks = [f"{year}-{m}-01" for m in range(1, 13)]
            xlabels = [calendar.month_abbr[m] for m in range(1, 13)]
            plt.xticks(ticks, xlabels)
        case "line":
            x_axis = []
            for date in dates:
                year, month, day = date.split("-")
                x_axis.append("-".join((year[-2:], month, day)))
            if show_median:
                plt.hline(median(weight_per_workout), "green")
            elif show_average:
                plt.hline(mean(weight_per_workout), "green")
            plt.date_form("y-m-d")
            plt.title("Weight Moved Per Workout")
            plt.xlabel("Date")
            plt.ylim(lower=0)
            plt.plotsize(90, 30)
            plt.plot(x_axis, weight_per_workout, marker="hd")
        case "bar":
            data = defaultdict(int)
            for date, weight in zip(dates, weight_per_workout):
                year, month, _ = date.split("-")
                month = calendar.month_abbr[int(month)]
                data["-".join((month, year[-2:]))] += weight
            if show_median:
                plt.vline(median(data.values()), "green")
            elif show_average:
                plt.vline(mean(data.values()), "green")
            plt.title("Weight Moved by Month")
            plt.xlabel("Weight Moved")
            plt.plotsize(90, 30)
            plt.bar(data.keys(), data.values(), orientation="h")
        case _:
            console.print(f":warning: {plot_type} not a valid plot.", style=WARNING)
            console.print(
                "Try running [underline]kettlebells stats --help[/underline] to see available plots",
                style=SUGGESTION,
            )
            return

    plt.theme("pro")
    console.print()
    plt.show()


def top_ten_workouts(data: dict, sort: str) -> Table:
    """Get the top ten workouts based on weight moved.

    Args:
        data: A dict of the data from the database.
        sort: A str of which parameter to sort the table by.

    Returns:
        Table: A rich Table of the top ten workouts.
    """
    units = data["loads"]["units"]
    workouts = []
    for workout_data in data["saved_workouts"]:
        date = workout_data["date"]
        workout = Workout(**workout_data["workout"])
        workouts.append((date, workout, workout.calc_workout_stats()))

    sort = sort.replace("-", " ")
    if sort == "time":
        workouts = sorted(workouts, key=lambda x: x[1].time, reverse=True)
    else:
        workouts = sorted(workouts, key=lambda x: x[2][sort], reverse=True)

    if len(workouts) > 10:
        workouts = workouts[:10]

    columns = [
        ("Date\n", "green"),
        ("Workout Type\n", "magenta"),
        ("Variation\n", "magenta"),
        ("Time\n(mins)", "magenta"),
        (f"Weight Moved\n({units})", "blue"),
        ("Reps\n", "blue"),
        (f"Weight Density\n({units}/min)", "blue"),
        ("Rep Density\n(reps/min)", "blue"),
    ]
    top_ten_table = Table(title=f"Top Ten Workouts by {sort.title()}")
    for col, style in columns:
        top_ten_table.add_column(col, style=style, justify="right")

    for date, workout, stats in workouts:
        top_ten_table.add_row(
            date,
            workout.workout_type.title(),
            workout.variation,
            f"{workout.time}",
            f"{stats['weight moved']:,}",
            f"{stats['reps']}",
            f"{stats['weight density']:.1f}",
            f"{stats['rep density']:.1f}",
        )
    return top_ten_table


def filter_by_program(data: dict) -> tuple[dict, str]:
    """Filter the workouts by program.
    Args:
        data: A dict of the data from the database.

    Returns: A tuple of the filtered data and the program name.
    """
    programs = {w["workout"]["workout_type"] for w in data["saved_workouts"]}
    program = iterfzf(
        programs,
        multi=False,
    )
    workouts = []
    for workout_data in data["saved_workouts"]:
        workout_type = workout_data["workout"]["workout_type"]
        if workout_type == program:
            workouts.append(workout_data)
    data["saved_workouts"] = workouts
    return data, program


def view_program(data: dict, program: str, display_workout: bool) -> Table:
    """Create a table based on a program.

    Args:
        data: A dict of the data from the database.
        program: The name of a workout program.

    Returns: A rich Table of the workouts from a program.
    """
    units = data["loads"]["units"]
    workouts = []
    for workout_data in data["saved_workouts"]:
        date = workout_data["date"]
        workout = Workout(**workout_data["workout"])
        load = [exercise.load for exercise in workout.exercises]
        if len(set(load)) == 1:
            load = load[0]
        else:
            load = "Varied"
        workouts.append((date, workout, load, workout.calc_workout_stats()))
        if display_workout:
            workout.display_workout()

    columns = [
        ("Date\n", "green"),
        ("Workout Type\n", "magenta"),
        ("Variation\n", "magenta"),
        ("Time\n(mins)", "magenta"),
        (f"load\n({units})", "blue"),
        (f"Weight Moved\n({units})", "blue"),
        ("Reps\n", "blue"),
        (f"Weight Density\n({units}/min)", "blue"),
        ("Rep Density\n(reps/min)", "blue"),
    ]
    program_table = Table(title=program.title())
    for col, style in columns:
        program_table.add_column(col, style=style, justify="right")

    for date, workout, load, stats in workouts:
        program_table.add_row(
            date,
            workout.workout_type.title(),
            workout.variation,
            f"{workout.time}",
            f"{load}",
            f"{stats['weight moved']:,}",
            f"{stats['reps']}",
            f"{stats['weight density']:.1f}",
            f"{stats['rep density']:.1f}",
        )
    return program_table


def retrieve_workout(data: dict, preview: bool) -> tuple[str, Workout]:
    """Retrieve a workout from the database.

    Args:
        data: A dict of the data from the database.
        preview: A bool of whether to use preview or not.

    Returns: A tuple of the date and the workout.
    """
    data = {w["date"]: Workout(**w["workout"]) for w in data["saved_workouts"]}
    if preview:
        date = iterfzf(
            data.keys(),
            multi=False,
            preview="rg {} " + str(KETTLEBELLS_HOME) + """ -A 20 -I """,
        )
    else:
        date = iterfzf(
            data.keys(),
            multi=False,
        )
    if date:
        workout = data[date]
        return date, workout
    sys.exit()


def table_to_df(rich_table: Table) -> pd.DataFrame:
    """Convert a rich.Table obj into a pandas.DataFrame obj with any rich formatting removed from the values.

    Args:
        rich_table (Table): A rich Table that should be populated by the DataFrame values.

    Returns:
        DataFrame: A pandas DataFrame with the Table data as its values."""

    table_data = {
        x.header.strip(): [Text.from_markup(y).plain for y in x.cells]
        for x in rich_table.columns
    }
    return pd.DataFrame(table_data)


def print_calendar(data: dict, year: int | None) -> None:
    """Highlight workout days and print a yearly calendar.

    Args:
        data: The database dictionary.
        year: The year of the calendar to print.

    """
    cal = calendar.Calendar()
    workout_dates = _get_dates(data)
    month_tables = []
    for month in range(1, 13):
        table = Table(
            title=f"{calendar.month_name[month]} {year}",
            style="blue",
            box=box.SIMPLE_HEAVY,
            padding=0,
        )
        for week_day in cal.iterweekdays():
            table.add_column(
                "{:.3}".format(calendar.day_name[week_day]), justify="right"
            )
        month_days = cal.monthdayscalendar(year, month)
        for weekdays in month_days:
            days = []
            for index, day in enumerate(weekdays):
                day_label = Text(str(day or ""))
                if day and (day, month, year) in workout_dates:
                    day_label.stylize("green")
                days.append(day_label)
            table.add_row(*days)
        month_tables.append(Align.center(table))

    columns = Columns(month_tables, padding=1, width=30)
    console.print()
    console.rule(str(year), style="magenta")
    console.print()
    console.print(columns, justify="center")
    console.rule(style="magenta")
    console.print()


def _get_dates(data: dict) -> list[tuple[int]]:
    """Get the workout dates from the database.
    Args:
        data: The database dictionary.

    returns:
        list[tuple[int]]: A list of tuples of the workout dates.
    """
    workout_dates = []
    for workout in data["saved_workouts"]:
        workout_date = datetime.strptime(workout["date"], DATE_FORMAT)
        workout_dates.append((workout_date.day, workout_date.month, workout_date.year))
    return workout_dates


def get_exercises_by_movement(data: dict, start: str = "", end: str = ""):
    """Create a table of exercises by the fundamental human movements.

    Args:
        data: The database dictionary.
        start: The start date of the workouts to include in the table.
        end: The end date of the workouts to include in the table.

    """
    units = data["loads"]["units"]
    exercises = data["exercises"]
    human_movements_reps = defaultdict(int)
    if start:
        start_date = datetime.strptime(start, DATE_FORMAT)
    if end:
        end_date = datetime.strptime(end, DATE_FORMAT)
    for workout_data in data["saved_workouts"]:
        workout = Workout(**workout_data["workout"])
        date_obj = datetime.strptime(workout_data["date"], DATE_FORMAT)
        if start and date_obj < start_date:
            continue
        if end and date_obj > end_date:
            continue
        for exercise in workout.exercises:
            if exercise.name in exercises:
                for movement in exercises[exercise.name]:
                    reps = exercise.sets * exercise.reps
                    human_movements_reps[movement] += reps
            else:
                print(f"Unknown exercise: {exercise.name}")
    columns = [
        ("Fundamental\nMovement", "magenta"),
        ("Reps\n", "blue"),
        (f"Weight Moved\n({units})", "blue"),
    ]
    movement_table = Table(title="Fundamental Human Movements")
    for col, style in columns:
        movement_table.add_column(col, style=style, justify="right")

    total_reps = sum(human_movements_reps.values())
    for movement, reps in sorted(
        human_movements_reps.items(), key=lambda item: item[1], reverse=True
    ):
        movement_table.add_row(
            movement.title(),
            f"{reps:,}",
            f"{reps / total_reps:.2%}",
        )
    movement_table.add_section()
    lower_body = human_movements_reps["squat"] + human_movements_reps["hinge"]
    upper_body = human_movements_reps["push"] + human_movements_reps["pull"]
    movement_table.add_row(
        "Lower Body", f"{lower_body:,}", f"{lower_body / total_reps:.2%}"
    )
    movement_table.add_row(
        "Upper Body", f"{upper_body:,}", f"{upper_body / total_reps:.2%}"
    )
    return movement_table



"""constants for kettlebells."""

from pathlib import Path

WARNING = "red"
SUGGESTION = "yellow"
FZF_DEFAULT_OPTS = "--height 13 --layout=reverse --border rounded --margin=2%,5%,10%,2%"
KETTLEBELLS_HOME = Path().home() / ".kettlebells"
KETTLEBELLS_DB = KETTLEBELLS_HOME / "kettlebells_db.json"
DATE_FORMAT = "%Y-%m-%d"
POUNDS_TO_KILOS_RATE = 0.45359237

IRON_CARDIO_PARAMS = {
    "bells": {"Single Bell": 4 / 6, "Double Bells": 2 / 6},
    "doublebell variations": {
        "Double Classic": 3 / 6,
        "Double Classic + Pullup": 2 / 6,
        "Double Traveling 2s": 1 / 6,
    },
    "singlebell variations": {
        "Classic": 6 / 12,
        "Classic + Pullup": 3 / 12,
        "Classic + Snatch": 2 / 12,
        "Traveling 2s": 1 / 12,
    },
    "times": {
        30: 1 / 6,
        20: 4 / 6,
        10: 1 / 6,
    },
    "loads": {
        "heavy load": 2 / 6,
        "medium load": 3 / 6,
        "light load": 1 / 6,
    },
    "swings": {True: 2 / 6, False: 4 / 6},
    "exercises": {
        "Classic": [
            ("Kettlebell Clean", 1),
            ("Kettlebell Press", 1),
            ("Kettlebell Front Squat", 1),
        ],
        "Classic + Pullup": [
            ("Kettlebell Clean", 1),
            ("Kettlebell Press", 1),
            ("Kettlebell Front Squat", 1),
            ("Pullup", 1),
        ],
        "Classic + Snatch": [
            ("Kettlebell Clean", 1),
            ("Kettlebell Press", 1),
            ("Kettlebell Front Squat", 1),
            ("Kettlebell Snatch", 1),
        ],
        "Traveling 2s": [
            ("Kettlebell Clean", 1),
            ("Kettlebell Press", 1),
            ("Kettlebell Front Squat", 1),
        ],
        "Double Classic": [
            ("Double Kettlebell Clean", 1),
            ("Double Kettlebell Press", 1),
            ("Double Kettlebell Front Squat", 1),
        ],
        "Double Classic + Pullup": [
            ("Double Kettlebell Clean", 1),
            ("Double Kettlebell Press", 1),
            ("Double Kettlebell Front Squat", 1),
            ("Pullup", 1),
        ],
        "Double Traveling 2s": [
            ("Double Kettlebell Clean", 1),
            ("Double Kettlebell Press", 1),
            ("Double Kettlebell Front Squat", 1),
        ],
    },
}

ABC_PARAMS = {
    "bells": {"Single Bell": 2 / 6, "Double Bells": 4 / 6},
    "doublebell variations": {"Armor Building Complex": 1},
    "singlebell variations": {"Armor Building Complex 2.0": 1},
    "times": {
        30: 1 / 10,
        25: 3 / 10,
        20: 3 / 10,
        15: 2 / 10,
        10: 1 / 10,
    },
    "loads": {
        "heavy load": 2 / 6,
        "medium load": 3 / 6,
        "light load": 1 / 6,
    },
    "swings": {True: 1 / 6, False: 5 / 6},
    "exercises": {
        "Armor Building Complex 2.0": [
            ("Kettlebell Clean", 2),
            ("Kettlebell Press", 2),
            ("Kettlebell Front Squat", 2),
        ],
        "Armor Building Complex": [
            ("Double Kettlebell Clean", 2),
            ("Double Kettlebell Press", 1),
            ("Double Kettlebell Front Squat", 3),
        ],
    },
}

BTB_PARAMS = {
    "2 C&P Ladders + Snatch": {
        "exercises": [
            {
                "name": "Kettlebell Clean and Press",
                "sets": 2,
                "reps": 20,
            },
            {
                "name": "Kettlebell Snatch",
                "sets": 1,
                "reps": 100,
            },
        ],
    },
    "3 C&P Ladders + Snatch": {
        "exercises": [
            {
                "name": "Kettlebell Clean and Press",
                "sets": 3,
                "reps": 20,
            },
            {
                "name": "Kettlebell Snatch",
                "sets": 1,
                "reps": 80,
            },
        ],
    },
    "5 C&P Ladders + Snatch": {
        "exercises": [
            {
                "name": "Kettlebell Clean and Press",
                "sets": 5,
                "reps": 20,
            },
            {
                "name": "Kettlebell Snatch",
                "sets": 1,
                "reps": 60,
            },
        ],
    },
    "2 C&P Ladders + Squats": {
        "exercises": [
            {
                "name": "Kettlebell Clean and Press",
                "sets": 2,
                "reps": 20,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "sets": 2,
                "reps": 10,
            },
        ],
    },
    "3 C&P Ladders + Squats": {
        "exercises": [
            {
                "name": "Kettlebell Clean and Press",
                "sets": 3,
                "reps": 20,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "sets": 3,
                "reps": 10,
            },
        ],
    },
    "5 C&P Ladders + Squats": {
        "exercises": [
            {
                "name": "Kettlebell Clean and Press",
                "sets": 5,
                "reps": 20,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "sets": 5,
                "reps": 10,
            },
        ],
    },
}

PW_PARAMS = {
    "Original": {
        "exercises": [
            {
                "name": "Half-kneeling Kettlebell Press",
                "sets": 3,
                "reps": 16,
            },
            {
                "name": "Hanging Leg Raise",
                "sets": 3,
                "reps": 8,
                "load": 0,
            },
            {
                "name": "Banded Hip Thrust",
                "sets": 3,
                "reps": 15,
            },
            {
                "name": "Bulgarian Goat Bag Swing",
                "sets": 3,
                "reps": 8,
            },
            {
                "name": "Goblet Squat",
                "sets": 1,
                "reps": 8,
            },
            {
                "name": "Broomstick Overhead Squat",
                "sets": 1,
                "reps": 8,
                "load": 0,
            },
        ],
    },
    "Indoor": {
        "exercises": [
            {
                "name": "Half-kneeling Kettlebell Press",
                "sets": 3,
                "reps": 16,
            },
            {
                "name": "Hanging Leg Raise",
                "sets": 3,
                "reps": 8,
                "load": 0,
            },
            {
                "name": "Hip Thrust",
                "sets": 1,
                "reps": 120,
            },
            {
                "name": "Clam Shell",
                "sets": 1,
                "reps": 120,
            },
            {
                "name": "Goblet Squat",
                "sets": 1,
                "reps": 8,
            },
            {
                "name": "Broomstick Overhead Squat",
                "sets": 1,
                "reps": 8,
                "load": 0,
            },
        ],
    },
    "The Bull": {
        "exercises": [
            {
                "name": "Half-kneeling Kettlebell Press",
                "sets": 3,
                "reps": 16,
            },
            {
                "name": "Hanging Leg Raise",
                "sets": 3,
                "reps": 8,
                "load": 0,
            },
            {
                "name": "Bulgarian Goat Bag Swing",
                "sets": 3,
                "reps": 8,
            },
            {
                "name": "Kettlebell Bent Over Row",
                "sets": 3,
                "reps": 16,
            },
            {
                "name": "Goblet Squat",
                "sets": 1,
                "reps": 8,
            },
            {
                "name": "Broomstick Overhead Squat",
                "sets": 1,
                "reps": 8,
                "load": 0,
            },
        ],
    },
}

THE_GIANT_PARAMS = {
    "The Giant 1.0": {
        "W1D1": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 5,
            },
        ],
        "W1D2": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 6,
            },
        ],
        "W1D3": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 4,
            },
        ],
        "W2D1": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 5,
            },
        ],
        "W2D2": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 6,
            },
        ],
        "W2D3": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 4,
            },
        ],
        "W3D1": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 5,
            },
        ],
        "W3D2": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 6,
            },
        ],
        "W3D3": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 4,
            },
        ],
        "W4D1": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 5,
            },
        ],
        "W4D2": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 6,
            },
        ],
        "W4D3": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 4,
            },
        ],
    },
}

THE_WOLF_PARAMS = {
    "The Wolf": {
        "W1D1": [
            {
                "name": "Kettlebell Snatch",
                "load": 0,
                "sets": 3,
                "reps": 10,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Press",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
        ],
        "W1D2": [
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Press",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
        ],
        "W1D3": [
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Press",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
        ],
        "W2D1": [
            {
                "name": "Kettlebell Snatch",
                "load": 0,
                "sets": 4,
                "reps": 10,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Press",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
        ],
        "W2D2": [
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Press",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
        ],
        "W2D3": [
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Press",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
        ],
        "W3D1": [
            {
                "name": "Kettlebell Snatch",
                "load": 0,
                "sets": 5,
                "reps": 10,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Press",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
        ],
        "W3D2": [
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Press",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
        ],
        "W3D3": [
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Press",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
        ],
        "W4D1": [
            {
                "name": "Kettlebell Snatch",
                "load": 0,
                "sets": 3,
                "reps": 10,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Press",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
        ],
        "W4D2": [
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Press",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
        ],
        "W4D3": [
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Press",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 3,
                "reps": 5,
            },
        ],
        "W5D1": [
            {
                "name": "Kettlebell Snatch",
                "load": 0,
                "sets": 4,
                "reps": 10,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Press",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
        ],
        "W5D2": [
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Press",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
        ],
        "W5D3": [
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Press",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 4,
                "reps": 5,
            },
        ],
        "W6D1": [
            {
                "name": "Kettlebell Snatch",
                "load": 0,
                "sets": 5,
                "reps": 10,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Press",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
        ],
        "W6D2": [
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Press",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
        ],
        "W6D3": [
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Press",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Clean",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 5,
                "reps": 5,
            },
        ],
    },
}

WORKOUT_GENERATOR_PARAMS = {
    "3 X 8": {
        "sets": 3,
    },
    "3 X 10": {
        "sets": 3,
    },
    "3 X 12": {
        "sets": 3,
    },
    "30/30 For 30": {
        "time": 30,
        "sets": 6,
    },
}

ABFB_PARAMS = {
    "variations": ["Program One", "Program Two", "Program Three"],
    "exercises": {
        "Clean and Press": {
            "reps": 8,
            "sets": 5,
        },
        "Curl": {
            "reps": 8,
            "sets": 5,
        },
    },
}

ABF_PARAMS = {
    "Intro": None,
    "Armor Building Complex": ABC_PARAMS["exercises"]["Armor Building Complex"],
    "Double Kettlebell Press": [
        ("Double Kettlebell Press", 2),
        ("Double Kettlebell Press", 3),
        ("Double Kettlebell Press", 5),
        ("Double Kettlebell Press", 10),
    ],
}


EASY_STRENGTH_PARAMS = {
    "regular": {
        "exercises": [
            {
                "name": "Deadlift",
                "sets": 3,
                "reps": 3,
                "load": 102,
            },
            {
                "name": "Double Kettlebell Press",
                "sets": 5,
                "reps": 2,
                "load": 32,
            },
            {
                "name": "Chinup",
                "sets": 3,
                "reps": 3,
                "load": 0,
            },
            {
                "name": "Ab Wheel",
                "sets": 1,
                "reps": 10,
                "load": 0,
            },
        ],
    },
    "light": {
        "exercises": [
            {
                "name": "Deadlift",
                "sets": 1,
                "reps": 10,
                "load": 86,
            },
            {
                "name": "Double Kettlebell Press",
                "sets": 1,
                "reps": 10,
                "load": 24,
            },
            {
                "name": "TRX I",
                "sets": 1,
                "reps": 5,
                "load": 0,
            },
            {
                "name": "TRX Y",
                "sets": 1,
                "reps": 5,
                "load": 0,
            },
            {
                "name": "TRX T",
                "sets": 1,
                "reps": 5,
                "load": 0,
            },
            {
                "name": "TRX High Pull",
                "sets": 1,
                "reps": 5,
                "load": 0,
            },
            {
                "name": "TRX Row",
                "sets": 1,
                "reps": 5,
                "load": 0,
            },
            {
                "name": "Ab Wheel",
                "sets": 1,
                "reps": 5,
                "load": 0,
            },
            {
                "name": "Hanging Leg Raise",
                "sets": 1,
                "reps": 5,
                "load": 0,
            },
        ],
    },
}

ROP_PARAMS = {
    "Heavy": {
        "exercises": [
            "Kettlebell Clean and Press",
            "Pullup",
            "Kettlebell Swing",
            "Done",
        ],
    },
    "Medium": {
        "exercises": [
            "Kettlebell Clean and Press",
            "Pullup",
            "Kettlebell Swing",
            "Done",
        ],
    },
    "Light": {
        "exercises": [
            "Kettlebell Clean and Press",
            "Pullup",
            "Kettlebell Snatch",
            "Done",
        ],
    },
}

DFW_PARAMS = {
    "dry fighting weight": {
        "W1D1": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": [1, 2, 3],
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 0,
                "reps": [1, 2, 3],
            },
        ],
        "W1D2": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 1,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 0,
                "reps": 1,
            },
        ],
        "W1D3": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 2,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 0,
                "reps": 2,
            },
        ],
        "W2D1": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": [1, 2, 3],
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 0,
                "reps": [1, 2, 3],
            },
        ],
        "W2D2": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 1,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 0,
                "reps": 1,
            },
        ],
        "W2D3": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 3,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 0,
                "reps": 3,
            },
        ],
        "W3D1": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": [1, 2, 3, 4],
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 0,
                "reps": [1, 2, 3, 4],
            },
        ],
        "W3D2": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 2,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 0,
                "reps": 2,
            },
        ],
        "W3D3": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 3,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 0,
                "reps": 3,
            },
        ],
        "W4D1": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": [1, 2, 3, 4, 5],
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 0,
                "reps": [1, 2, 3, 4, 5],
            },
        ],
        "W4D2": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 2,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 0,
                "reps": 2,
            },
        ],
        "W4D3": [
            {
                "name": "Double Kettlebell Clean and Press",
                "load": 0,
                "sets": 0,
                "reps": 2,
            },
            {
                "name": "Double Kettlebell Front Squat",
                "load": 0,
                "sets": 0,
                "reps": 2,
            },
        ],
    },
}

EXERCISES = {
    "Kettlebell Press": ["push"],
    "Kettlebell Clean and Press": ["hinge", "push", "pull"],
    "Kettlebell Clean": ["hinge", "pull"],
    "Goblet Squat": ["squat"],
    "Swing": ["hinge"],
    "Kettlebell Swing": ["hinge"],
    "Ab Wheel": ["core"],
    "Arch Body Leglift": ["core"],
    "Plank Shoulder Tap": ["core"],
    "Banded Hip Thrust": ["hinge"],
    "Bent Over Row": ["pull"],
    "Kettlebell Bent Over Row": ["pull"],
    "Hands Elevated Pike Pushup": ["push"],
    "Arch Body Leg Lift": ["core"],
    "Side Plank Lift": ["core"],
    "Bent Press": ["hinge", "push"],
    "Bird Dog": ["core"],
    "Bodyweight Row": ["pull"],
    "Bodyweight Squat": ["squat"],
    "Broomstick Overhead Squat": ["squat"],
    "Bulgarian Goat Bag Swing": ["hinge"],
    "Chinup": ["pull"],
    "Clean and Press": ["hinge", "push", "pull"],
    "Clean": ["hinge", "pull"],
    "Curl": ["upper body isolation"],
    "Lateral Raise": ["upper body isolation"],
    "Shrimp Squat": ["squat"],
    "Overhead Squat": ["squat"],
    "Overhead Box Squat": ["squat"],
    "TRX Shrimp Squat": ["squat"],
    "TRX Jump Squat": ["squat"],
    "One Arm Swing": ["hinge"],
    "Deadlift": ["hinge"],
    "Dip": ["push"],
    "Anchor Squat": ["squat"],
    "Done": [None],
    "Double Clean and Press": ["hinge", "push", "pull"],
    "Double Bent Over Row": ["pull"],
    "Double Clean": ["hinge", "pull"],
    "Double Deadlift": ["hinge"],
    "Double Front Squat": ["squat"],
    "Double Jerk": ["push"],
    "Double Press": ["push"],
    "Double Push Press": ["push"],
    "Double Split Squat": ["squat"],
    "Double Swing": ["hinge"],
    "Double Kettlebell Clean and Press": ["hinge", "push", "pull"],
    "Double Kettlebell Bent Over Row": ["pull"],
    "Double Kettlebell Clean": ["hinge"],
    "Double Kettlebell Deadlift": ["hinge"],
    "Double Kettlebell Front Squat": ["squat"],
    "Double Kettlebell Jerk": ["push"],
    "Double Kettlebell Press": ["push"],
    "Double Kettlebell Push Press": ["push"],
    "Double Kettlebell Split Squat": ["squat"],
    "Double Kettlebell Swing": ["hinge"],
    "Racked Kettlebell Squat": ["squat"],
    "Front Squat": ["squat"],
    "Half-kneeling Kettlebell Press": ["push"],
    "Half-kneeling Press": ["push"],
    "Hanging Leg Raise": ["core"],
    "Hip Thrust": ["hinge"],
    "Hollow Body Rock": ["core"],
    "Jerk": ["push"],
    "Lateral Squat": ["squat"],
    "Other": [None],
    "Pike Pushup": ["push"],
    "Pistol Squat": ["squat"],
    "Pullup": ["pull"],
    "Press": ["push"],
    "Push Press": ["push"],
    "Pushup Position Plank": ["core"],
    "Pushup": ["push"],
    "Renegade Row": ["pull"],
    "Scapular Push Up": ["core"],
    "Side Plank Hip Lift": ["core"],
    "Side Plank": ["core"],
    "Side Lift": ["core"],
    "Single Leg Deadlift": ["hinge"],
    "Kettlebell Snatch": ["hinge", "pull"],
    "Snatch": ["hinge", "pull"],
    "Split Squat": ["squat"],
    "TRX Advanced Row": ["pull"],
    "TRX Bicep Curl": ["upper body isolation"],
    "TRX I": ["pull"],
    "TRX Pushup": ["push"],
    "TRX Row": ["pull"],
    "TRX Single Arm Row": ["pull"],
    "TRX T": ["pull"],
    "TRX Y": ["pull"],
    "TRX Fly": ["upper body isolation"],
    "Thruster": ["squat", "push"],
    "Toes to Bar": ["core"],
    "Turkish Get-up": ["squat", "core"],
    "Windmill": ["core"],
    "Leg Press": ["squat"],
}

HUMAN_MOVEMENTS = [
    "push",
    "pull",
    "squat",
    "hinge",
    "core",
    "upper body isolation",
    "lower body isolation",
    "done",
]

BODYWEIGHT_FACTORS = {
    "Bodyweight Row": 0.5,
    "Chinup": 0.96,
    "Dip": 0.96,
    "Pike Pushup": 0.64,
    "Pullup": 0.96,
    "Pushup": 0.64,
    "TRX Advanced Row": 0.64,
    "TRX Pushup": 0.5,
    "TRX Row": 0.3,
}



from pathlib import Path
from unittest import mock

import kettlebells.__main__
import pytest
from kettlebells.__init__ import __version__
from kettlebells.__main__ import cli
from typer.testing import CliRunner

from .test_constants import TEST_IC_WORKOUT

runner = CliRunner()


def test_version():
    result = runner.invoke(cli, ["--version"])
    assert result.exit_code == 0
    assert f"{__version__}\n" in result.stdout


@mock.patch("kettlebells.__main__.Confirm.ask")
def test_save_without_save(confirm_mock, database):
    """Test workout not saved renders correctly."""
    confirm_mock.return_value = ""
    with mock.patch.object(kettlebells.__main__, "KETTLEBELLS_DB", Path(database.name)):
        result = runner.invoke(cli, ["save"])
        assert "Last workout generated:" in result.stdout
        assert "Workout not saved." in result.stdout


@pytest.mark.parametrize(
    "func, option",
    [
        ("create_ic_or_abc", "ic"),
        ("create_btb_workout", "btb"),
        ("create_custom_workout", "custom"),
        ("create_perfect_workout", "pw"),
        ("create_time_based_workout", "giant"),
        # ("from_dict", None),
    ],
)
def test_matching(func, option):
    with mock.patch("kettlebells.__main__." + func) as save_mock:
        runner.invoke(cli, ["save", "--workout-type", option])
        save_mock.assert_called_once()


def test_save_bad_input(database):
    """Test workout not saved renders correctly."""
    with mock.patch.object(kettlebells.__main__, "KETTLEBELLS_DB", Path(database.name)):
        result = runner.invoke(cli, ["save", "-w asdf"])
        assert "'asdf' is not an option" in result.stdout


@mock.patch("kettlebells.__main__.Confirm.ask")
@mock.patch("kettlebells.__main__.Prompt.ask")
@mock.patch("kettlebells.__main__.save_workout")
def test_save(save_mock, prompt_mock, confirm_mock, database):
    """Test the save command prints out renders the last workout generated and the stats."""
    confirm_mock.return_value = True
    prompt_mock.side_effect = ["2023-09-12"]

    with mock.patch.object(kettlebells.__main__, "KETTLEBELLS_DB", Path(database.name)):
        result = runner.invoke(cli, ["save"])
        assert "Workout Stats" in result.stdout
        save_mock.assert_called_once()


def test_last(database):
    """Test the last command reads the last saved session and displays the date and stats."""
    with mock.patch.object(kettlebells.__main__, "KETTLEBELLS_DB", Path(database.name)):
        result = runner.invoke(cli, ["last"])
        assert "Date: Sep 14, 2023" in result.stdout


@mock.patch("kettlebells.stats.iterfzf")
def test_view(iterfzf_mock, database):
    """Test the view command."""
    iterfzf_mock.return_value = "2023-09-14"
    with mock.patch.object(kettlebells.__main__, "KETTLEBELLS_DB", Path(database.name)):
        result = runner.invoke(cli, ["view"])
        assert "Date: Sep 14, 2023" in result.stdout


@mock.patch("kettlebells.__main__.filter_by_program")
def test_view_program(filter_mock, database):
    """Test the view -program command."""
    filter_mock.return_value = "2023-09-14"
    with mock.patch.object(kettlebells.__main__, "KETTLEBELLS_DB", Path(database.name)):
        runner.invoke(cli, ["view", "-P"])
        filter_mock.assert_called_once()


@mock.patch("kettlebells.__main__.set_loads")
@mock.patch("kettlebells.__main__.read_database")
@mock.patch("kettlebells.__main__.write_database")
def test_setloads(write_mock, read_mock, loads_mock):
    """Test setloads command calls set_loads, read_database, and write_database."""
    runner.invoke(cli, ["setloads"])
    loads_mock.assert_called_once()
    read_mock.assert_called_once()
    write_mock.assert_called_once()


@mock.patch("kettlebells.__main__.initialize_database")
def test_init(init_mock):
    """Test init command calls initialize_database."""
    result = runner.invoke(cli, ["init"])
    assert result.exit_code == 0
    init_mock.assert_called_once()


@mock.patch("kettlebells.__main__.top_ten_workouts")
def test_stats_best(best_mock, database):
    """Test stats command with best flag calls read_database and top_ten_workouts."""
    with mock.patch.object(kettlebells.__main__, "KETTLEBELLS_DB", Path(database.name)):
        result = runner.invoke(cli, ["stats", "-b"])
        assert result.exit_code == 0
        best_mock.assert_called_once()


@mock.patch("kettlebells.__main__.plot_workouts")
def test_stats_plot(plot_mock, database):
    """Test stats command with plot flag."""
    with mock.patch.object(kettlebells.__main__, "KETTLEBELLS_DB", Path(database.name)):
        result = runner.invoke(cli, ["stats", "-p", "line"])
        assert result.exit_code == 0
        plot_mock.assert_called_once()


@mock.patch("kettlebells.__main__.print_calendar")
def test_stats_calendar(cal_mock, database):
    """Test stats command with calendar flag."""
    with mock.patch.object(kettlebells.__main__, "KETTLEBELLS_DB", Path(database.name)):
        result = runner.invoke(cli, ["stats", "-c", "-y", "2023"])
        assert result.exit_code == 0
        cal_mock.assert_called_once()


@mock.patch("kettlebells.__main__.read_database")
@mock.patch("kettlebells.__main__.get_all_stats")
def test_stats(stats_mock, read_mock):
    """Test getting the stats"""
    runner.invoke(cli, ["stats"])
    read_mock.assert_called_once()
    stats_mock.assert_called_once()


@mock.patch("kettlebells.__main__.confirm_loads")
@mock.patch("kettlebells.__main__.random_ic_or_abc")
@mock.patch("kettlebells.__main__.cache_workout")
def test_random(cache_mock, workout_mock, confirm_mock):
    """Test getting the stats"""
    workout_mock.return_value = TEST_IC_WORKOUT
    runner.invoke(cli, ["random", "-w ic"])
    workout_mock.assert_called_once()
    cache_mock.assert_called_once()
    confirm_mock.assert_called_once()



from kettlebells.workouts import Exercise, Workout

TEST_IC_WORKOUT = Workout(
    **{
        "bodyweight": 90,
        "units": "kg",
        "variation": "Double Classic + Pullup",
        "time": 30,
        "exercises": [
            Exercise(
                **{
                    "name": "Double Kettlebell Clean",
                    "load": 28,
                    "sets": 20,
                    "reps": 1,
                },
            ),
            Exercise(
                **{
                    "name": "Double Kettlebell Press",
                    "load": 28,
                    "sets": 20,
                    "reps": 1,
                },
            ),
            Exercise(
                **{
                    "name": "Double Kettlebell Front Squat",
                    "load": 28,
                    "sets": 20,
                    "reps": 1,
                },
            ),
            Exercise(
                **{
                    "name": "Pullup",
                    "load": 86,
                    "sets": 20,
                    "reps": 1,
                },
            ),
            Exercise(
                **{
                    "name": "Kettlebell Swing",
                    "load": 28,
                    "sets": 1,
                    "reps": 60,
                },
            ),
        ],
        "workout_type": "iron cardio",
    }
)

TEST_SINGLE_TRAVELING_2S_WORKOUT = Workout(
    **{
        "bodyweight": 90,
        "units": "kg",
        "variation": "Traveling 2s",
        "time": 10,
        "exercises": [
            Exercise(
                **{
                    "name": "Kettlebell Clean",
                    "load": 20,
                    "sets": 13,
                    "reps": 1,
                },
            ),
            Exercise(
                **{
                    "name": "Kettlebell Press",
                    "load": 20,
                    "sets": 13,
                    "reps": 1,
                },
            ),
            Exercise(
                **{
                    "name": "Kettlebell Front Squat",
                    "load": 20,
                    "sets": 13,
                    "reps": 1,
                },
            ),
            Exercise(
                **{
                    "name": "Kettlebell Swing",
                    "load": 20,
                    "sets": 1,
                    "reps": 100,
                },
            ),
        ],
        "workout_type": "iron cardio",
    }
)
TEST_DOUBLE_TRAVELING_2S_WORKOUT = Workout(
    **{
        "bodyweight": 90,
        "units": "kg",
        "variation": "Double Traveling 2s",
        "time": 29,
        "exercises": [
            Exercise(
                **{
                    "name": "Double Kettlebell Clean",
                    "load": 28,
                    "sets": 16,
                    "reps": 1,
                },
            ),
            Exercise(
                **{
                    "name": "Double Kettlebell Press",
                    "load": 28,
                    "sets": 16,
                    "reps": 1,
                },
            ),
            Exercise(
                **{
                    "name": "Double Kettlebell Front Squat",
                    "load": 28,
                    "sets": 16,
                    "reps": 1,
                },
            ),
            Exercise(
                **{
                    "name": "Kettlebell Swing",
                    "load": 28,
                    "sets": 1,
                    "reps": 50,
                },
            ),
        ],
        "workout_type": "iron cardio",
    }
)

TEST_BTB_WORKOUT = Workout(
    **{
        "bodyweight": 90,
        "units": "kg",
        "variation": "2 C&P Ladders + Snatch",
        "time": 30,
        "exercises": [
            Exercise(
                **{
                    "name": "Kettlebell Clean and Press",
                    "load": 24,
                    "sets": 2,
                    "reps": 20,
                }
            ),
            Exercise(
                **{"name": "Kettlebell Snatch", "load": 20, "sets": 1, "reps": 100}
            ),
        ],
        "workout_type": "back to basics",
    },
)
TEST_PERFECT_WORKOUT = Workout(
    **{
        "bodyweight": 90,
        "units": "kg",
        "variation": "The Bull",
        "time": 10,
        "exercises": [
            Exercise(
                **{
                    "name": "Half-kneeling Kettlebell Press",
                    "load": 20,
                    "sets": 3,
                    "reps": 16,
                },
            ),
            Exercise(
                **{
                    "name": "Hanging Leg Raise",
                    "load": 0,
                    "sets": 3,
                    "reps": 8,
                },
            ),
            Exercise(
                **{
                    "name": "Bulgarian Goat Bag Swing",
                    "load": 24,
                    "sets": 3,
                    "reps": 8,
                },
            ),
            Exercise(
                **{
                    "name": "Kettlebell Bent Over Row",
                    "load": 20,
                    "sets": 3,
                    "reps": 16,
                },
            ),
            Exercise(
                **{
                    "name": "Goblet Squat",
                    "load": 20,
                    "sets": 1,
                    "reps": 8,
                },
            ),
            Exercise(
                **{
                    "name": "Broomstick Overhead Squat",
                    "load": 0,
                    "sets": 1,
                    "reps": 8,
                },
            ),
        ],
        "workout_type": "perfect workout",
    },
)
TEST_CUSTOM_WORKOUT = Workout(
    **{
        "bodyweight": 90,
        "units": "kg",
        "variation": "Custom",
        "time": 30,
        "exercises": [
            Exercise(**{"name": "Turkish Get-up", "load": 24, "sets": 1, "reps": 6}),
            Exercise(**{"name": "TRX T", "load": 0, "sets": 3, "reps": 8}),
        ],
        "workout_type": "custom",
    },
)
TEST_GIANT_WORKOUT = Workout(
    **{
        "bodyweight": 90,
        "units": "kg",
        "variation": "W1D1",
        "workout_type": "The Giant 1.0",
        "time": 30,
        "exercises": [
            Exercise(
                **{
                    "name": "Double Kettlebell Clean and Press",
                    "load": 24,
                    "sets": 10,
                    "reps": 5,
                }
            ),
        ],
    },
)

TEST_WOLF_WORKOUT = Workout(
    **{
        "bodyweight": 90,
        "units": "kg",
        "variation": "W1D1",
        "workout_type": "The Wolf",
        "time": 12,
        "exercises": [
            Exercise(
                **{
                    "name": "Kettlebell Snatch",
                    "load": 16,
                    "sets": 3,
                    "reps": 10,
                }
            ),
            Exercise(
                **{
                    "name": "Double Kettlebell Front Squat",
                    "load": 16,
                    "sets": 3,
                    "reps": 5,
                }
            ),
            Exercise(
                **{
                    "name": "Double Kettlebell Press",
                    "load": 16,
                    "sets": 3,
                    "reps": 5,
                }
            ),
            Exercise(
                **{
                    "name": "Double Kettlebell Front Squat",
                    "load": 16,
                    "sets": 3,
                    "reps": 5,
                }
            ),
            Exercise(
                **{
                    "name": "Double Kettlebell Clean",
                    "load": 16,
                    "sets": 3,
                    "reps": 5,
                }
            ),
            Exercise(
                **{
                    "name": "Double Kettlebell Front Squat",
                    "load": 16,
                    "sets": 3,
                    "reps": 5,
                }
            ),
        ],
    },
)


TEST_ABFB = Workout(
    **{
        "bodyweight": 90,
        "units": "kg",
        "variation": "Program Three",
        "time": 32,
        "exercises": [
            Exercise(
                **{
                    "name": "Clean and Press",
                    "load": 20,
                    "sets": 1,
                    "reps": 8,
                },
            ),
            Exercise(
                **{
                    "name": "Front Squat",
                    "load": 20,
                    "sets": 1,
                    "reps": 1,
                },
            ),
            Exercise(
                **{
                    "name": "Clean and Press",
                    "load": 29,
                    "sets": 1,
                    "reps": 8,
                },
            ),
            Exercise(
                **{
                    "name": "Front Squat",
                    "load": 29,
                    "sets": 1,
                    "reps": 1,
                },
            ),
            Exercise(
                **{
                    "name": "Clean and Press",
                    "load": 43,
                    "sets": 3,
                    "reps": 8,
                },
            ),
            Exercise(
                **{
                    "name": "Front Squat",
                    "load": 43,
                    "sets": 3,
                    "reps": 1,
                },
            ),
            Exercise(
                **{
                    "name": "Curl",
                    "load": 20,
                    "sets": 2,
                    "reps": 8,
                },
            ),
            Exercise(
                **{
                    "name": "Curl",
                    "load": 25,
                    "sets": 3,
                    "reps": 8,
                },
            ),
        ],
        "workout_type": "armor building formula barbell",
    }
)


TEST_WORKOUT_NO_SWINGS = Workout(
    **{
        "bodyweight": 90,
        "units": "kg",
        "variation": "Classic + Snatch",
        "time": 20,
        "exercises": [
            Exercise(
                **{
                    "name": "Kettlebell Clean",
                    "load": 24,
                    "sets": 16,
                    "reps": 1,
                },
            ),
            Exercise(
                **{
                    "name": "Kettlebell Press",
                    "load": 24,
                    "sets": 16,
                    "reps": 1,
                },
            ),
            Exercise(
                **{
                    "name": "Kettlebell Front Squat",
                    "load": 24,
                    "sets": 16,
                    "reps": 1,
                },
            ),
            Exercise(
                **{
                    "name": "Kettlebell Snatch",
                    "load": 24,
                    "sets": 16,
                    "reps": 1,
                },
            ),
        ],
        "workout_type": "iron cardio",
    }
)

TEST_WORKOUT_SINGLE_BELL_PULLUPS = Workout(
    **{
        "bodyweight": 90,
        "units": "kg",
        "variation": "Classic + Pullup",
        "time": 10,
        "exercises": [
            Exercise(
                **{
                    "name": "Kettlebell Clean",
                    "load": 20,
                    "sets": 10,
                    "reps": 1,
                },
            ),
            Exercise(
                **{
                    "name": "Kettlebell Press",
                    "load": 20,
                    "sets": 10,
                    "reps": 1,
                },
            ),
            Exercise(
                **{
                    "name": "Kettlebell Front Squat",
                    "load": 20,
                    "sets": 10,
                    "reps": 1,
                },
            ),
            Exercise(
                **{
                    "name": "Pullup",
                    "load": 90,
                    "sets": 5,
                    "reps": 1,
                },
            ),
        ],
        "workout_type": "iron cardio",
    }
)

TEST_WORKOUT_GENERATOR_WORKOUT = Workout(
    **{
        "bodyweight": 90,
        "units": "kg",
        "variation": "3 X 10",
        "time": 40,
        "exercises": [
            Exercise(
                **{
                    "name": "Deadlift",
                    "sets": 3,
                    "reps": 10,
                    "load": 100,
                },
            ),
            Exercise(
                **{
                    "name": "Bench Press",
                    "sets": 3,
                    "reps": 10,
                    "load": 80,
                },
            ),
        ],
        "workout_type": "workout generator",
    }
)
TEST_EASY_STRENGTH_WORKOUT = Workout(
    **{
        "bodyweight": 90,
        "units": "kg",
        "variation": "24",
        "time": 30,
        "exercises": [
            Exercise(
                **{
                    "name": "Goblet Squat",
                    "sets": 1,
                    "reps": 10,
                    "load": 32,
                },
            ),
            Exercise(
                **{
                    "name": "Kettlebell Swing",
                    "sets": 5,
                    "reps": 15,
                    "load": 32,
                },
            ),
            Exercise(
                **{
                    "name": "Dip",
                    "sets": 2,
                    "reps": 5,
                    "load": 98,
                },
            ),
            Exercise(
                **{
                    "name": "Double Kettlebell Front Squat",
                    "sets": 2,
                    "reps": 5,
                    "load": 28,
                },
            ),
            Exercise(
                **{
                    "name": "Kettlebell Bent Over Row",
                    "sets": 2,
                    "reps": 10,
                    "load": 32,
                },
            ),
            Exercise(
                **{
                    "name": "Ab Wheel",
                    "sets": 1,
                    "reps": 10,
                    "load": 0,
                },
            ),
        ],
        "workout_type": "easy strength",
    }
)

TEST_ABF_PRESS_WORKOUT = Workout(
    **{
        "bodyweight": 90,
        "units": "kg",
        "variation": "Double Kettlebell Press",
        "time": 30,
        "exercises": [
            Exercise(
                **{
                    "name": "Double Kettlebell Press",
                    "sets": 5,
                    "reps": 2,
                    "load": 20,
                },
            ),
            Exercise(
                **{
                    "name": "Double Kettlebell Press",
                    "sets": 5,
                    "reps": 3,
                    "load": 20,
                },
            ),
            Exercise(
                **{
                    "name": "Double Kettlebell Press",
                    "sets": 5,
                    "reps": 5,
                    "load": 20,
                },
            ),
            Exercise(
                **{
                    "name": "Double Kettlebell Press",
                    "sets": 5,
                    "reps": 10,
                    "load": 20,
                },
            ),
        ],
        "workout_type": "armor building formula",
    }
)
TEST_ABF_ABC_WORKOUT = Workout(
    **{
        "bodyweight": 90,
        "units": "kg",
        "variation": "Armor Building Complex",
        "time": 30,
        "exercises": [
            Exercise(
                **{
                    "name": "Double Kettlebell Clean",
                    "sets": 30,
                    "reps": 2,
                    "load": 20,
                },
            ),
            Exercise(
                **{
                    "name": "Double Kettlebell Press",
                    "sets": 30,
                    "reps": 1,
                    "load": 20,
                },
            ),
            Exercise(
                **{
                    "name": "Double Kettlebell Front Squat",
                    "sets": 30,
                    "reps": 3,
                    "load": 20,
                },
            ),
        ],
        "workout_type": "armor building formula",
    }
)
TEST_ROP_WORKOUT = Workout(
    **{
        "bodyweight": 90,
        "units": "kg",
        "variation": "Medium",
        "time": 45,
        "exercises": [
            Exercise(
                **{
                    "name": "Kettlebell Clean and Press",
                    "sets": 5,
                    "reps": 12,
                    "load": 28,
                },
            ),
            Exercise(
                **{
                    "name": "Pullup",
                    "sets": 5,
                    "reps": 6,
                    "load": 86,
                },
            ),
            Exercise(
                **{
                    "name": "Kettlebell Swing",
                    "sets": 5,
                    "reps": 10,
                    "load": 28,
                },
            ),
        ],
        "workout_type": "rite of passage",
    }
)
TEST_DATA = {
    "loads": {
        "units": "kg",
        "bodyweight": 90,
        "light load": 20,
        "medium load": 24,
        "heavy load": 28,
        "the giant": 24,
        "the wolf": 16,
    },
    "exercises": {
        "Kettlebell Press": ["push"],
        "Goblet Squat": ["squat"],
    },
    "saved_workouts": [
        {
            "date": "2023-09-12",
            "workout": {
                "bodyweight": 90,
                "units": "kg",
                "variation": "W1D1",
                "time": 20,
                "exercises": [
                    {
                        "name": "Double Kettlebell Clean and Press",
                        "load": 24,
                        "sets": 8,
                        "reps": 3,
                    },
                    {
                        "name": "Double Kettlebell Front Squat",
                        "load": 24,
                        "sets": 8,
                        "reps": 3,
                    },
                ],
                "workout_type": "Dry Fighting Weight",
            },
        },
        {
            "date": "2023-09-14",
            "workout": {
                "bodyweight": 90,
                "units": "kg",
                "variation": "Double Classic",
                "time": 20,
                "exercises": [
                    {
                        "name": "Double Kettlebell Clean",
                        "load": 20,
                        "sets": 29,
                        "reps": 1,
                    },
                    {
                        "name": "Double Kettlebell Press",
                        "load": 20,
                        "sets": 29,
                        "reps": 1,
                    },
                    {
                        "name": "Double Kettlebell Front Squat",
                        "load": 20,
                        "sets": 29,
                        "reps": 1,
                    },
                ],
                "workout_type": "iron cardio",
            },
        },
    ],
    "cached_workouts": [
        {
            "bodyweight": 90,
            "units": "kg",
            "variation": "Double Classic + Pullup",
            "time": 30,
            "exercises": [
                {
                    "name": "Double Kettlebell Clean",
                    "load": 28,
                    "sets": 20,
                    "reps": 1,
                },
                {
                    "name": "Double Kettlebell Press",
                    "load": 28,
                    "sets": 20,
                    "reps": 1,
                },
                {
                    "name": "Double Kettlebell Front Squat",
                    "load": 28,
                    "sets": 20,
                    "reps": 1,
                },
                {
                    "name": "Pullup",
                    "load": 90,
                    "sets": 20,
                    "reps": 1,
                },
                {
                    "name": "Kettlebell Swing",
                    "load": 28,
                    "sets": 6,
                    "reps": 10,
                },
            ],
            "workout_type": "iron cardio",
        }
    ],
}

TEST_NEW_EXERCISES = {
    "leg press": ["squat"],
    "front squat": ["squat"],
}
